/* x11 -- Fortran callable ARKPLOT driver for X Windows software 

Started life as the PGPLOT driver for workstations running X Windows.

  Version 1.0 - 1989 Nov 06 - A. L. Fey
                              Initial try.
  Version 1.1 - 1990 Feb 15 - A. L. Fey
                              Add 'line of pixels'
                              and 'area fill'.
  Version 2.0 - 1990 Jun 07 - A. L. Fey
                              Merge functionality with S.C.
                              Allendorf's Fortran XEDRIVER.FOR.
                              Add code to implement a backing store.
  Version 2.1 - 1990 Jun 28 - A. L. Fey
                              Fix-up display visual classification.
  Version 2.2 - 1990 Jul 06 - A. L. Fey
                              Add additional code for alternate
                              implementation of a backing store.
                              This method 'forks' a process to run
                              in the background and thus requires
                              more overhead than the 'signal' method.
  Version 2.3 - 1990 Oct  8 - Jim Trice (trice@asta.pa.uky.edu)
                              Fix opcode 1.
  Version 2.4 - 1991 Mar 19 - Jim Morgan (morgan@astro.umd.edu)
                              Fix cursor problems; signals.
                1991 Apr 30 - SNS/CIT Trivial modifications for sharable 
                              libraries
                1991 Oct 14 - TJP/CIT Corrections to strings 

  ARKPLOT version July 1993 - Timn and Koji.

*/

#define SIGNAL 0                  /* Choose implementation of backing */
#define FORK   1                  /* store. Choose one by setting it
                                     to 1. The SIGNAL method seems to
                                     work well under Berkeley based Unix 
                                     while the FORK method works under
                                     both Berkeley and System V Unix.
                                     If no backing store is required,
                                     set both to 0 */
                                     
                                  /* Get includes */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/keysym.h>
#include <stdio.h>
#include <sys/time.h>
#include <signal.h>

                                  /* Define some constants */

#define TRUE  1
#define FALSE 0
#define BELL  7
#define COLORMULT 65534
#define NCOLORS 8
#define MAXPOINTS 1000 /* this is sloppy! */

                                  /* Useful macros */

#ifndef min
#define min(a,b) ((a)<(b)?(a):(b))
#endif

#ifndef max
#define max(a,b) ((a)>(b)?(a):(b))
#endif

                                  /* Global variables */

static Display *display;
static Window window;
static GC gc;
static Pixmap pixmap;

                                  /* Begin x11 */

void x11_(inchar, real1, real2, real3, real4, integ, ifunc)

int *integ, *ifunc;
float *real1, *real2, *real3, *real4;
char  *inchar;


{

                                  /* X structures */

      static Colormap cmap;
      static Visual *visual;
      static Window parent;
      static GC gcb;
      static Pixmap icon_pixmap;
      static Cursor cursor;
      static XColor colorcell_defs[NCOLORS];
      static XColor c_black, c_red, c_green;
      static XSizeHints size_hints;
      static XEvent report;
      static XGCValues values;
      static XSetWindowAttributes setwinattr;
      static XWMHints wmhints;
      static XImage *xi;
      static XPoint points[MAXPOINTS];
      static KeySym keysym;
      static XComposeStatus compose;
      static XVisualInfo VisualInfo;

                                  /* Window variables */

      static int cells;
      static int screen;
      static unsigned int width;
      static unsigned int height;
      static unsigned long value_mask;
      static int depth;
      static int x, y;
      static unsigned int border_width = 4;
      static unsigned int display_width, display_height;
      static unsigned int display_widthMM, display_heightMM;
      char *window_name = "ARK graphics";
      char *icon_name = "ARKplot";
      static int cursor_cross = 34;  /* XC_crosshair */

                                  /* PGPLOT color table (RGB) */

      static float ctable[NCOLORS][3] =
         { {0.0,0.0,0.0}, {1.0,1.0,1.0}, {1.0,0.0,0.0}, {0.0,1.0,0.0},
           {0.0,0.0,1.0}, {0.0,1.0,1.0}, {1.0,0.0,1.0}, {1.0,1.0,0.0} };

                                  /* Various variables */

      static unsigned long plane_masks[NCOLORS];
      static unsigned long pixels[NCOLORS];
      static unsigned long black, white, color;
      static char buffer[10];
      static unsigned char image[1][1280];
      static float resol[2];
      static int maxcol;
      static int icount = 0, npoints;
      static int imin, imax, jmin, jmax;
      static int xmin, xmax, ymin, ymax;
      static int i0, i1, j0, j1, ic;
      static float factor;
      static int mono, Static;
      static int i;
      static int first = 0;
      static int icurs, jcurs;

#if FORK
      static int running = FALSE, pid;
#endif

#if SIGNAL
                                  /* Timer for Expose event handler */

      static int running = FALSE;
      static struct itimerval ovalue, tvalue = {
         {0, 100000},                     /* 0.1 second interval */
         {0, 100000}                      /* 0.1 second value */
         };
      static grxw03();                    /* Expose event handler */
#endif

/* Turn OFF Expose event handler before we do anything */

#if SIGNAL
         signal (SIGALRM, SIG_IGN);
#endif

#if FORK
         if (running)
            kill (pid, SIGSTOP);
#endif

/* Branch on opcode. */

         switch (*ifunc) {

/*--- IFUNC=9, Open workstation -----------------------------------------*/

         case 9 :

              if (first != 0) {
                /* Clear the pixmap */
                XFillRectangle (display, pixmap, gcb, 0, 0, width, height);
                /* Clear the window */
                XClearWindow (display, window); }
              else
                {
                first = 1;

         /* Connect to X server */

              if ((display = XOpenDisplay ('\0')) == NULL) {
                 (void) fprintf (stderr,
                    "XWDRIV: cannot connect to X server\n");
                 break;
                 }

         /* Get screen size from display structure macro */

              screen = DefaultScreen (display);

         /* Size and position window */

              display_width  = DisplayWidth (display, screen);
              display_height = DisplayHeight (display, screen);
              display_widthMM  = DisplayWidthMM (display, screen);
              display_heightMM = DisplayHeightMM (display, screen);

              resol[0] = 25.4 * ((float) display_width /
                                 (float) display_widthMM);
              resol[1] = 25.4 * ((float) display_height /
                                 (float) display_heightMM);

              factor = 9.0 / 13.0;
              width = 7 * display_width / 10;
              height = factor * width;

              factor = 105.0 / 100.0;
              x = display_width - factor * width;
              y = display_height - factor * height;

              imin = (int) (0.25 * resol[0] + 0.5);
              jmin = (int) (0.25 * resol[1] + 0.5);
              imax = width - imin - 1;
              jmax = height - jmin - 1;

              *real1 = (float) (imax-imin);
              *real2 = (float) (jmax-jmin);
              *real3 = 0.0;
              *real4 = 0.0;
              *integ = 0;

         /* Classify the display and create a color map */

              black = BlackPixel (display, screen);
              white = WhitePixel (display, screen);

              depth  = DisplayPlanes (display, screen);
              visual = DefaultVisual (display, screen);
              parent = RootWindow    (display, screen);

              if (depth == 1) {
                 /* Revert to monochrome */
                 mono = TRUE;
                 }
              else {
                 switch (visual->class) { 

                 case PseudoColor :
                      /* Get the default color map */
                      cmap = DefaultColormap (display, screen);
                      value_mask = 0;
                      mono = FALSE;
                      Static = FALSE;
                      break;
                 case StaticColor :
                      /* Get the default color map */
                      cmap = DefaultColormap (display, screen);
                      value_mask = 0;
                      mono = FALSE;
                      Static = TRUE;
                      break;
                 default :
                      /* Default visual is not one we can use;
                         try to find one of type PseudoColor */
                      if (!XMatchVisualInfo (display, screen, depth,
                             PseudoColor, &VisualInfo)) {
                         /* Revert to monochrome */
                         mono = TRUE;
                         }
                      else {
                         /* Found visual of type PseudoColor */
                         visual = VisualInfo.visual;
                         /* Create a color map */
                         cmap = XCreateColormap (display, parent, visual,
                            AllocNone);
                         setwinattr.colormap = cmap;
                         value_mask = CWColormap;
                         mono = FALSE;
                         Static = FALSE;
                         }
                      break;
                      }
                 }

         /* Create a window */

              if (!mono)
                 window = XCreateWindow (display, parent,
                    x, y, width, height, border_width, depth,
                    InputOutput, visual, value_mask, &setwinattr);
              else
                 window = XCreateSimpleWindow (display, parent,
                    x, y, width, height, border_width,
                    white, black);

         /* Set window manager hints to assure keyboard input */

              wmhints.flags = InputHint;
              wmhints.input = FALSE;
              XSetWMHints (display, window, &wmhints);

         /* Set the maximum number of colors and allocate color cells
            if the color map is read/write */

              maxcol = 1;  /* Default for monochrome device */

              if (!mono) {

                 /* Determine the maximum number of colors available */
                 cells  = DisplayCells (display, screen);
                 maxcol = min (cells, NCOLORS);

                 /* Grab as many color cells as we need (or can) */ 
                 if (!Static) {
                    while (!XAllocColorCells (display, cmap, True,
                            plane_masks, 0, pixels, (unsigned) maxcol) &&
                            maxcol > 2) {
                       maxcol--;
                       }
                    }

                 maxcol -= 1;

                 /* Revert to monochrome if two or fewer colors were found */ 
                 if (maxcol <= 1) {
                    mono = TRUE;
                    maxcol = 1;
                    }

                 }

         /* Load the color table */

              if (!mono) {

                 /* Store PGPLOT color definitions */
                 for (i = 0; i <= min (NCOLORS - 1, maxcol); i++) {
                    colorcell_defs[i].pixel = pixels[i];
                    colorcell_defs[i].red   = (int)(ctable[i][0]*COLORMULT+0.5);
                    colorcell_defs[i].green = (int)(ctable[i][1]*COLORMULT+0.5);
                    colorcell_defs[i].blue  = (int)(ctable[i][2]*COLORMULT+0.5);
                    colorcell_defs[i].flags = DoRed | DoGreen | DoBlue;
                    colorcell_defs[i].pad   = 0;

                    if (!Static) {
                       XStoreColor (display, cmap, &colorcell_defs[i]);
                       }
                    else {
                       XAllocColor (display, cmap, &colorcell_defs[i]);
                       pixels[i] = colorcell_defs[i].pixel;
                       }
                    }

                 /* Redefine black and white */
                 black = pixels[0];
                 white = pixels[1];

                 /* Get color structures for cursor colors */
                 c_black.pixel = pixels[0];
                 c_red.pixel   = pixels[2];
                 c_green.pixel = pixels[3];
                 XQueryColor (display, cmap, &c_black);
                 XQueryColor (display, cmap, &c_red);
                 XQueryColor (display, cmap, &c_green);

                 }

         /* Set the window colors */

              XSetWindowBackground (display, window, black);
              XSetWindowBorder (display, window, white);

         /* Initialize size hint property for window manager */

              size_hints.flags = PPosition | PSize | PMinSize | PMaxSize;
              size_hints.x = x;
              size_hints.y = y;
              size_hints.width = width;
              size_hints.height = height;
              size_hints.min_width = width;
              size_hints.min_height = height;
              size_hints.max_width = width;
              size_hints.max_height = height;

         /* Set properties for window manager (always before mapping) */

              XSetStandardProperties (display, window, window_name,
                 icon_name, icon_pixmap, NULL, 0, &size_hints);

         /* Create a pixmap for double buffering */

              pixmap = XCreatePixmap (display, window, width, height, depth);

         /* Create default graphics contexts and make a few adjustments */

              gc  = XCreateGC (display, pixmap, 0, &values);
              gcb = XCreateGC (display, pixmap, 0, &values);
              XSetFillRule  (display, gc, WindingRule);
              XSetFillStyle (display, gc, FillSolid);
              XSetFillRule  (display, gcb, WindingRule);
              XSetFillStyle (display, gcb, FillSolid);
  
         /* Specify foreground colors in graphics contexts */

              XSetForeground (display, gc, white);
              XSetForeground (display, gcb, black);

         /* Clear the pixmap - we do this for servers that do not
            clear the pixmap when it is created */

              XFillRectangle (display, pixmap, gcb, 0, 0, width, height);

         /* Select the event types wanted */

              XSelectInput (display, window, ExposureMask);
   
         /* Display window */

              XMapWindow (display, window);

         /* Wait for mapping notification */

              XNextEvent (display, &report);

#if SIGNAL
         /* Setup timer for Expose event handler */

              setitimer (0, &tvalue, &ovalue);
              running = TRUE;
#endif

#if FORK
         /* Start Expose event handler */

              if (!running) {
                 pid = fork ();
                 if (pid == 0) {
                    grxw04 (); 
                    }
                 running = TRUE;
                 }
#endif

         /* Initialize the damaged region */

              grxw02 (width, height, &xmin, &xmax, &ymin, &ymax);

         /* Successful-- return */

/* Set up first position that cursor will appear at. */

              icurs = (imax + imin)/2.0;
              jcurs = (jmax + jmin)/2.0;

                }

              break;

/*--- IFUNC=10, Close workstation ---------------------------------------*/

         case 10 :

#if FORK
         /* Kill Expose event handler before we do anything */

              if (running) {
                 kill (pid, SIGKILL);
                 running = FALSE;
                 }
#endif
#if SIGNAL
              running = FALSE;
#endif

              XUnmapWindow (display, window);
              XFreeGC (display, gc);
              XFreeGC (display, gcb);
              XDestroyWindow (display, window);
              XFreePixmap (display, pixmap);
              XCloseDisplay (display);
              first = 0;
              break;

/*--- IFUNC=12, Draw line -----------------------------------------------*/

         case 12 :

         /* Translate input */

              i0 = (int) (*real1 + 0.5) + imin;
              j0 = (jmax - jmin) - (int) (*real2 + 0.5) + jmin;
              i1 = (int) (*real3 + 0.5) + imin;
              j1 = (jmax - jmin) - (int) (*real4 + 0.5) + jmin;

         /* Draw the line */

              XDrawLine (display, pixmap, gc, i0, j0, i1, j1);

         /* Update the damaged region */

              grxw01 (1, i0, j0, i1, j1, &xmin, &xmax, &ymin, &ymax);

              break;


/*--- IFUNC=15, Select color index --------------------------------------*/

         case 15 :

              ic = *integ;

         /* Check input for proper range */

              if (ic < 0 || ic > maxcol) ic = 1;

         /* Change the color index - handle monochrome properly */

              if (!mono) {
                 color = pixels[ic];
                 XSetForeground (display, gc, color);
                 }
              else if (ic == 1)
                 XSetForeground (display, gc, white);
              else
                 XSetForeground (display, gc, black);

              break;

/*--- IFUNC=16, Flush buffer. -------------------------------------------*/

         case 16 :

         /* Copy pixmap to window and flush display */

              if (xmax != -1)
                 XCopyArea (display, pixmap, window, gc, xmin, ymin,
                       xmax - xmin + 1, ymax - ymin + 1, xmin, ymin);
              XFlush (display);

         /* Reset the damaged region */

              grxw02 (width, height, &xmin, &xmax, &ymin, &ymax);

         /* Pop the window down to the bottom 

              XLowerWindow(display, window);  */

              break;

/*--- IFUNC=17, Read cursor. --------------------------------------------*/

         case 17 :

         /* First make sure the display is up to date. */
         XFlush (display);

         /* Create graphics cursor */

              cursor = XCreateFontCursor (display, cursor_cross);
              XDefineCursor (display, window, cursor);
              if (!mono)
                 XRecolorCursor (display, cursor, &c_red, &c_black);

         /* Move cursor to wherever it was last used */

              XWarpPointer (display, None, window, icurs, jcurs,
                 width, height, icurs, jcurs);

              XSelectInput (display, window, StructureNotifyMask
                 | ExposureMask | ButtonPressMask
                 | EnterWindowMask | LeaveWindowMask);

         /* Discard any ButtonPress or KeyPress events encountered
            up till now */

              while (XCheckTypedEvent (display, ButtonPress, &report))
                 ;
              while (XCheckTypedEvent (display, KeyPress, &report))
                 ;

         /* Event loop */

              while (1) {

                 XNextEvent (display, &report);
                 switch (report.type) {

                 case Expose :
                      XCopyArea (display, pixmap, window, gc,
                         report.xexpose.x, report.xexpose.y,
                         report.xexpose.width, report.xexpose.height,
                         report.xexpose.x, report.xexpose.y);
                      break;
                 case ButtonPress :
                      *real1 = (float) (report.xbutton.x - imin);    
                      *real2 = (float) ((jmax - jmin) - 
                                         report.xbutton.y + jmin);    
                      if (report.xbutton.button == Button1)
                         strcpy (inchar, " ");
                      else if (report.xbutton.button == Button2)
                         strcpy (inchar, "r");
                      else
                         strcpy (inchar, "e");
                         /* Squirrel away the cursor position for next time. */
                         icurs=report.xbutton.x;
                         jcurs=report.xbutton.y;
                      goto endcursor;
                 case EnterNotify :
                      XSetInputFocus (display, window,
                         RevertToPointerRoot, CurrentTime);
                      break;
                 case LeaveNotify :
                      XSetInputFocus (display, PointerRoot,
                         RevertToPointerRoot, CurrentTime);
                      break;
                 default :
                      break;
                      }
                 }

              endcursor :

         /* Return cursor to original state */

              /* Return to old input selection added by Timn. */
              XSelectInput(display, window, ExposureMask);
              /* Free up cursor added by Timn. */
              XUndefineCursor(display, window);
              XFreeCursor(display, cursor);

              break;



/*--- IFUNC=?, ----------------------------------------------------------*/

         default :

         /* Notify the user of an input error */

              (void) fprintf (stderr,
                 "Unimplemented function in X Windows device driver: %d\n",
                 *ifunc);
              break;

              } /* End of switch */

/* Turn ON Expose event handler, then return to calling program */

#if SIGNAL
         if (running)
             signal (SIGALRM, grxw03);
#endif

#if FORK
         if (running)
            kill (pid, SIGCONT);
#endif

         return;

} /* End of x11 */

/*GRXW01 -- PGPLOT XWindow driver, calculate 'damaged' region.
            From S.C. Allendorf's XEDRIVER.FOR */

grxw01 (line, i0, j0, i1, j1, xmin, xmax, ymin, ymax)
int line, i0, j0, i1, j1;
int *xmin, *xmax, *ymin, *ymax;
{
                                        /* Update the damaged region. */
      if (i0 > *xmax)
         *xmax = i0;
      if (i0 < *xmin)
         *xmin = i0;
      if (j0 > *ymax)
         *ymax = j0;
      if (j0 < *ymin)
         *ymin = j0;
                                        /* See if we were passed a
                                           rectangle and update the
                                           damaged region accordingly. */
      if (line == 1) {
         if (i1 > *xmax)
            *xmax = i1;
         if (i1 < *xmin)
            *xmin = i1;
         if (j1 > *ymax)
            *ymax = j1;
         if (j1 < *ymin)
            *ymin = j1;
         }
}

/*GRXW02 -- PGPLOT XWindow driver, reset 'damaged' region.
            From S.C. Allendorf's XEDRIVER.FOR */

grxw02 (width, height, xmin, xmax, ymin, ymax)
unsigned int width, height;
int *xmin, *xmax, *ymin, *ymax;
{
                                        /* Reset the boundaries of the
                                           damaged region. */
      *xmax = -1;
      *ymax = -1;
      *xmin = width + 1;
      *ymin = height + 1;
}

#if SIGNAL
/*GRXW03 -- PGPLOT XWindow driver, Expose event handler (redrawing routine).
            From S.C. Allendorf's XEDRIVER.FOR */

static grxw03 (sig, code, scp)
int sig, code;
struct sigcontext *scp;
{
      XEvent event;
      int event_mask;

                                       /* Select events. */

      event_mask = ExposureMask | EnterWindowMask | LeaveWindowMask;

      while (XCheckWindowEvent (display, window, event_mask, &event)) {

                                       /* If part of the window has been
                                          exposed, redraw that part. We
                                          ignore NoExpose events and 
                                          GraphicsExpose events. */

            switch (event.type) {

            case Expose :
                 XCopyArea (display, pixmap, window, gc,
                    event.xexpose.x, event.xexpose.y,
                    event.xexpose.width, event.xexpose.height,
                    event.xexpose.x, event.xexpose.y);
                 break;
            case EnterNotify :
                 XSetInputFocus (display, window,
                    RevertToPointerRoot, CurrentTime);
                 break;
            case LeaveNotify :
                 XSetInputFocus (display, PointerRoot,
                    RevertToPointerRoot, CurrentTime);
                 break;
            default :
                 break;
                 }
            }
}
#endif

#if FORK
/*GRXW04 -- PGPLOT XWindow driver, Expose event handler (redrawing routine).
            From S.C. Allendorf's XEDRIVER.FOR */

grxw04 ()
{
      XEvent event;

      while (1) {

                                       /* If part of the window has been
                                          exposed, redraw that part. We
                                          ignore NoExpose events and 
                                          GraphicsExpose events. */


         XNextEvent (display, &event);
         switch (event.type) {

         case Expose :
              XCopyArea (display, pixmap, window, gc,
                 event.xexpose.x, event.xexpose.y,
                 event.xexpose.width, event.xexpose.height,
                 event.xexpose.x, event.xexpose.y);
              break;
         case EnterNotify :
              XSetInputFocus (display, window,
                 RevertToPointerRoot, CurrentTime);
              break;
         case LeaveNotify :
              XSetInputFocus (display, PointerRoot,
                 RevertToPointerRoot, CurrentTime);
              break;
         default :
              break;
              }
         }
}
#endif
