    program bradley

      use ark_file_io
      use opt_extr
      use shape

      implicit none

      character(len=80) :: root

      ! The profile shape from the PSF star.
      real, dimension(6) :: a_par
      ! The clipping radius for the optimal photometry.
      real :: cliprad
      real :: skycnt, skyerr, skynos
      integer :: ibox
      real :: fwhm
      character :: cflag, sky_flag

      character(len=10) :: chdummy

      type(a_star), allocatable, dimension(:) :: star
      real, allocatable, dimension(:) :: flux_s, flux_e_s
      type(a_colour), allocatable, dimension(:) :: save_col
      type(a_star) :: test_star

      ! For the image.
      integer, dimension(2) :: naxis, low, high
      integer :: iflag, iext
      real, dimension(:,:), allocatable :: data, axdata
      character, dimension(:,:), allocatable :: pix_flg
      real :: gain

      integer :: nx, ny

      ! The number of stars, and counters.
      integer :: nstar, istar, jstar

      character(len=100) :: header
      integer :: iostat
      integer, parameter :: outfil=11, infil=10

      call setbug()

      print*, '> Give orders for polynomial.'
      read(*,*) nx, ny


      frame: do
        call extark('fit')
        ! Read in the data.  The image is contained in the array data.
        iflag=inpark(naxis, data, axdata)
        if (iflag < 0) exit frame
        ! And any flags.
        iflag=inpflg(naxis, pix_flg)
        ! Recover the name of the last file read by lstark.
        call lstark_in(root, iext)
        ! Remove leading blanks.
        root=adjustl(root)
        iflag=index(root, '.')
        if (iflag /= 0) root(iflag:len(root))=' '

        if (iext == 0) then
          ! First time around so write headers for files.
          print*, 'Opening the file ', trim(root)//'_opt.cat'
          open(unit=outfil, file=trim(root)//'_opt.cat', status='old')
          print*, 'Opening the file ', trim(root)//'_old_opt.cat'
          open(unit=infil, file=trim(root)//'_old_opt.cat')
          read(outfil,'(a100)') header
          write(infil,'(a100)') header
          read(outfil,'(a100)') header
          write(infil,'(a100)') header
          read(outfil,'(a100)') header
          write(infil,'(a100)') header
          ! And now copy al the data
          copy: do
            iostat=read_star(outfil, test_star, 1)
            if (iostat < 0) exit copy
            call write_star(infil, test_star, 1)
          end do copy
          close(infil)
          ! And now leave unit outfil read to be overwritten.
          rewind(outfil)
          read(outfil,'(/)')
          write(outfil,'(a100)') header
        else
          print*, 'Opening the file ', trim(root)//'_opt.cat'
          open(unit=outfil, file=trim(root)//'_opt.cat', &
          status='old', position='append')
        end if

        if (naxis(1)*naxis(2) > 0) then
          if ( & 
            get_header_i('ARK-LO1', low(1))  < 0 .or.   &
            get_header_i('ARK-LO2', low(2))  < 0 .or.   &
            get_header_i('ARK-HI1', high(1)) < 0 .or.  &
            get_header_i('ARK-HI2', high(2)) < 0) then
            print*, 'The active area needs setting for ', root
            print*, 'using the program pix_info.  Going to next file.'
            deallocate(data, axdata, pix_flg)
            cycle frame
          end if
          if (get_header_r('ARK-GAIN', gain) < 0) then
            print*, 'The gain needs setting for ', root
            print*, 'using the program pix_info.  Going to next file.'
            deallocate(data, axdata, pix_flg)
            cycle frame
          end if

!!!!! Needs replacement with read_cluster_file.
          ! Decode the header, first the seeing, and then the rotation, and 
          ! finally the clipping radius.
          ! I do this by removing everything but the numbers, and reading those.
          print*, 'Opening file ', trim(root)//'_old_opt.cat'
          open(unit=infil, file=trim(root)//'_old_opt.cat', action='read')
          read (infil,*) chdummy, chdummy, chdummy, chdummy, chdummy, chdummy, &
          a_par(1:3), cliprad
          fwhm=sqrt(a_par(1)*a_par(2))
          a_par(1)=a_par(1)/1.665
          read(infil,'(/)') 
          nstar=0
          count_them: do
            iostat=read_star(infil, test_star, 1)
            if (iostat /= 0) exit count_them
            if (iext /= test_star%ccd) cycle count_them
            nstar=nstar+1
          end do count_them
          rewind(infil)
          print*, 'There are ', nstar, ' stars for this extension.'

          allocate(star(nstar))
          call zero_star(star)
      
          ! Read in the data.
          read(infil, '(/,/)')
          istar=0
          do 
            iostat=read_star(infil, test_star, 1)
            if (iostat /= 0) exit
            if (iext /= test_star%ccd) cycle
            istar=istar+1
            star(istar)=test_star
          end do
          close(infil)

          ! In principle it would be better to sort the stars, but it
          ! means you have to re-order the other arrays as well.
          !write(*,*) '* Sorting in order of flux.'
          !call starsort(star(1:nstar))

          ! Before we start faffing with them, save the current "colour".
          allocate(save_col(nstar))
          save_col=star%col(1)

          
          write(*,*) '* Now measuring shapes.'
          allocate(flux_s(nstar), flux_e_s(nstar))
          ! Now measure the flux with a tighter mask.
          ! Reset Gaussian width.
          a_par(1:2)=a_par(1:2)/2.0
          a_par(4)=1.0
          jstar=0
          each_star: do istar=1, nstar
            !print*, istar, star(istar)%col(1)
            if (star(istar)%col(1)%flg /= 'OO') cycle each_star
            jstar=jstar+1
            if (1000*(jstar/1000) == jstar) then
              print*, 'Done ', jstar, ' stars.'
            end if
            ibox=int(sqrt(453.2*fwhm*fwhm+ 4.0*fwhm*fwhm))
            call skyfit(data, star(istar)%x, star(istar)%y, fwhm, ibox, low, &
            high, pix_flg, skycnt, skyerr, skynos, sky_flag)
            ! Set position for this star.
            a_par(5)=star(istar)%x
            a_par(6)=star(istar)%y
            flux_s(istar)=sum_flux(data, pix_flg, skycnt, skyerr, skynos, &
            a_par, 2.0*fwhm, 0.0, gain, low, high, flux_e_s(istar), cflag)
            ! flux_s(istar)=sum_aper(data, 3.0*fwhm, skycnt, skyerr, skynos, &
            ! star(istar)%x, star(istar)%y, gain, low, high, pix_flg, &
            ! flux_e_s(istar), cflag)
            if (sky_flag/='O' .and. cflag=='O') cflag=sky_flag
            if ('O'//cflag /= star(istar)%col(1)%flg) then
              ! If the flags arn't the same, make sure a bad one is taken.
              if (star(istar)%col(1)%flg=='OO') star(istar)%col(1)%flg='O'//cflag
            end if
            call mag_to_flux(star(istar)%col(1))
            !print*, star(istar)%col(1), flux_s(istar), flux_e_s(istar)
          end do each_star

          ! Now add the entension number onto root.
          write(root(len_trim(root)+1:len_trim(root)+3), '(a, i2.2)') &
          '_', iext
      
          if (nx>=0 .and. ny>=0) call halley(flux_s, flux_e_s, &
          star, nstar, nx, ny, root)
          deallocate(flux_s, flux_e_s)

          ! Put the old flags and mags back, except for stuff that's 
          ! gained a non-stellar flag.
          where(star%col(1)%flg /= 'ON') 
            star%col(1)=save_col
          elsewhere
            star%col(1)%data=save_col%data
            star%col(1)%err =save_col%err
          endwhere
          deallocate(save_col)
        
          ! Write out to the new opt file.
          do istar=1, nstar
            call write_star(outfil, star(istar), 1)
          end do

          deallocate(star)
          deallocate(data, axdata, pix_flg)

        end if

        close(outfil)

      end do frame

    end program bradley

         
