PROGRAM DIVIDE

!   Program to read and divide data files created by the ARK program OPPHOT.

    implicit none

    character(len=50) :: finput, foutput, fname
    character(len=50), allocatable, dimension(:) :: compfiles
    integer :: i, j, k, l, bad, eof1, flag
    integer :: datapts, ncomp, nunit
    integer :: badfile, resfile
    integer :: counter, good
    real :: eflag
    double precision :: time, counts, ecounts, eex
    double precision :: comptest1, comptest2
    double precision :: compcounts, comperror
    double precision, allocatable, dimension(:,:) :: data
    double precision, allocatable, dimension(:,:,:) :: comp

!   Set constants
    good = 0
    bad = 0
    i = 0
    j = 0
    datapts = 0
    counter = 0
    nunit = 10

!   Get the names of the files and open them
    print*
    print*, "Enter the name of the file to be divided:"
    read(*,'(a)') finput
    print*
    open (unit=nunit, file=finput, status='old', access='sequential')
    read (nunit,*)
    read (nunit,*)
    read (nunit,*)

!   Count the number of lines in the input file and allocate the data array.
    do
       read (nunit,*,iostat=eof1)time, counts, ecounts, eex, eflag
       if (eof1/=0) exit
       datapts=datapts+1
    end do
    print*, "Input file has",datapts,"data points"
    print*
    allocate(data(5,datapts))
    eof1=0
    close(nunit)

!   Ask for how many comparison stars and allocate the appropriate arrays.
    print*, "How many comparison stars do you want to use?"
    read*, ncomp
    print*
    allocate(compfiles(ncomp))

!   Go round this loop and make sure the comparison star files are ok.
    do j=1, ncomp

       nunit=nunit+10
       print*, "Enter name for comparison star",j
       read*, compfiles(j)
       fname = compfiles(j)
       open (unit=nunit, file=fname, status='old', access='sequential')
       read (nunit,*)
       read (nunit,*)
       read (nunit,*)
       do
         read (nunit,*,iostat=eof1)time, counts, ecounts, eex, eflag
         if (eof1/=0) exit
         counter = counter + 1
       end do

       if (counter/=datapts) then
          print*, "***WARNING: File",fname,"is not consistent with input file ***"
          print*, "*** Exiting .... "
          stop
       end if
 
       counter = 0
       eof1 = 0
       close(nunit)

    end do

    allocate(comp(ncomp,5,datapts))

!   Re-open the data files, read in the data and set up the bad.dat files.
    nunit = 10
    open (unit=nunit, file=finput, status='old', access='sequential')
    read (nunit,*)
    read (nunit,*)
    read (nunit,*)
    do j=1, datapts
       read(nunit,*,iostat=eof1)data(1,j), data(2,j), data(3,j), data(4,j), data(5,j)
       if (eof1/=0) exit
    end do
    eof1 = 0
    close(nunit)

    do j=1, ncomp

       fname = compfiles(j)
       nunit=nunit+10
       open (unit=nunit, file=fname, status='old', access='sequential')
       read (nunit,*)
       read (nunit,*)
       read (nunit,*)

       do i=1, datapts
         read (nunit,*,iostat=eof1)comp(j,1,i), comp(j,2,i), comp(j,3,i), comp(j,4,i), comp(j,5,i)
         if (eof1/=0) exit

!        Check for time mismatch in the files.
         if (j==1) then
            l=2
         else
            l=j
         end if
         if (comp(j,1,i)/=comp(l-1,1,i).or.comp(j,1,i)/=data(1,i)) then
            print*, '*** WARNING: File mismatch!'
            print*, data(1,i), comp(:,1,i)
            goto 10
         end if

       end do   

       eof1 = 0
       close(nunit)
       i = 0

    end do

    badfile = nunit + 10
    resfile = badfile + 10

    open (unit=badfile, file="bad.dat", status='unknown', access='sequential')
    write (badfile,*)'! These are the bad points found by the MULDIV program'
    write (badfile,*)'! The lines go as follows: the first one belonging to the'
    write (badfile,*)'! data input file and the following to the comparison star files 1, 2...'

!   Setup the result file.
    print*, "Give name for the output file:"
    print*
    read*, foutput
    open (unit=resfile, file=foutput, status='unknown', access='sequential')
    write (resfile,*)
    write (resfile,*)
    write (resfile,*)

!   Go round this loop and do the division.
    division: do k=1, datapts

        i=i+1

!       Check for bad points. If they exist store them in bad.dat.
        comptest1 = minval(comp(:,3,k))
        comptest2 = sum(comp(:,5,k))/dble(ncomp)
        if (comptest1<=0.0d0.or.comptest2/=0.or.data(5,k)<0.0d0) then
           flag=0
           call bad_points (data, comp, flag, bad, k, ncomp, badfile, datapts)
           cycle division
        end if

!       Do the division and error bar calculations.
        compcounts = (sum(comp(:,2,k)))/dble(ncomp)
        comperror = sqrt(sum(comp(:,3,k)**2.0d0))

        counts=(data(2,k)/compcounts)
        ecounts=(abs(counts*(sqrt(((data(3,k)/data(2,k))**2.0d0) &
                                    +((comperror/compcounts)**2.0d0)))))/dble(ncomp)

!       Check for points that produce bad results.
        if (counts<ecounts) then
           flag=0
           call bad_points (data, comp, flag, bad, k, ncomp, badfile, datapts)
           close(101)
           cycle division
        end if
         
        good=good+1

!       Write the output to file.
        write (resfile,*)data(1,k), real(counts), real(ecounts), data(4,k)

    end do division

!   Screen output.
    print*
    print*, "File Division Operation Completed"
    print*
    print*, "Files have       " ,i, "points"
    print*, "Output file has  " ,good, " points"
    print*, "Bad points found:", bad
    print*
    if (bad.eq.0) then
       close(unit=badfile, status='delete')
    else 
       close(unit=badfile)
       print*, "Bad points saved in bad.dat"
       print*
    end if

10 END PROGRAM DIVIDE

!---------------------------------------------------------------------------

subroutine bad_points (data, comp, flag, bad, k, ncomp, badfile, datapts)

    character(len=20) :: frame
    integer :: bad, flag, eof2
    integer :: k, l, ncomp, badfile, datapts
    double precision :: times, checkup, checkdown, eex
    double precision :: data(5,datapts)
    double precision :: comp(ncomp,5,datapts)

    bad = bad + 1

    open (unit=101, file='times.dat', status='old', access='sequential', err=20)

    bad_loop: do

        read(101,*,iostat=eof2)frame
        read(101,*,iostat=eof2)times, eex
        if (eof2/=0) exit

        checkup=times+eex+tiny(times)
        checkdown=times-eex-tiny(times)
        if (data(1,k).ge.checkdown.and.data(1,k).le.checkup) then
           write(badfile,*)frame
           write(badfile,*)real(data(1,k)), real(data(2,k)), real(data(3,k)), real(data(5,k))
           do l=1, ncomp
              write(badfile,*)real(comp(l,1,k)), real(comp(l,2,k)), real(comp(l,3,k)), real(comp(l,5,k))
              flag=1
           end do
        else
           cycle bad_loop
        end if

    end do bad_loop

20  if (flag.eq.1) return
    write(badfile,*)
    write(badfile,*)real(data(1,k)), real(data(2,k)), real(data(3,k)), real(data(5,k))
    do l=1, ncomp
       write(badfile,*)real(comp(l,1,k)), real(comp(l,2,k)), real(comp(l,3,k)), real(comp(l,5,k))
    end do
    return

end subroutine bad_points
