     module shape 
      
       use var_apcor_mod
       use define_star
       use opt_extr
       
       implicit none
      
       contains
       
       subroutine shape_filter(data, star, nstar, seeing, &
       low, high, noise_map, gain, pix_flg, nx, ny, ifname, sort)

         ! Measures the shapes of stars, and then flags those which are
         ! not point-like.

         real, dimension(:,:), intent(inout) :: data
         type(a_star), dimension(:), intent(inout) :: star
         integer, intent(in) :: nstar
         real, intent(in) :: seeing
         integer, dimension(:), intent(in) :: low, high
         real, dimension(:,:), intent(in) :: noise_map
         real, intent(in) :: gain
         ! The pixel flags.
         character, dimension(:,:), intent(in) :: pix_flg
         integer, intent(in) :: nx, ny
         character(len=*), intent(in) :: ifname
         logical, intent(in) :: sort

         ! For fitting the position, and determining the flux.
         real, dimension(6) :: a_par
         real, dimension(2) :: e_pos
         real, allocatable, dimension(:) :: flux_s, flux_e_s
         !real, allocatable, dimension(:) :: flux_b, flux_e_b
         real :: noise
         integer :: i
         character :: cflag_fit, cflag
         
         write(*,*) '* Measuring fluxes.'
         do i=1, nstar
           if (1000*(i/1000) == i) then
             print*, 'Doing ', i, ' out of ', nstar, ' stars.'
             !print*, 'At ', star(i)%x, star(i)%y
           end if
           ! Set up for fitting the star.
           ! Set Gaussian width.
           a_par(1)=seeing/1.665
           a_par(2)=seeing/1.665
           ! And the rotation (in radians).
           a_par(3)=0.0
           ! Set position.
           a_par(5)=star(i)%x
           a_par(6)=star(i)%y
           if (nint(star(i)%x)<low(1) .or. nint(star(i)%x)>high(1) .or. &
               nint(star(i)%y)<low(2) .or. nint(star(i)%y)>high(2) ) then
             star(i)%col(1)%flg='OE'
           else  
             ! Set noise.
             noise=noise_map(nint(star(i)%x), nint(star(i)%y))
             ! Set normalisation.
             a_par(4)=data(nint(star(i)%x), nint(star(i)%y))
             call gfit(data, .true., 6.0, .false., 0.0, &
             noise, gain, low, high, pix_flg, a_par, e_pos, cflag_fit)
             ! Update the star's position.
             star(i)%x=a_par(5)
             star(i)%y=a_par(6)
             ! An a_par of zero will cause a crash in sum_flux.
             if (abs(a_par(4)) < 10.0*tiny(a_par(4))) &
             a_par(4)=10.0*tiny(a_par(4))
             ! Normal optimal extraction.
             star(i)%col(1)%data=sum_flux(data, pix_flg, 0.0, 0.0, noise, &
             a_par, 2.0*seeing, 0.0, gain, low, high, star(i)%col(1)%err, &
             cflag)
             if (cflag_fit/='O' .and. cflag=='O') cflag=cflag_fit
             star(i)%col(1)%flg='O'//cflag
           end if
         end do

         if (sort) then
           write(*,*) '* Sorting in order of flux.'
           star%col(1)%data=-star%col(1)%data
           call sort_star(star, nstar, 'increasing_col(1)%data')
           star%col(1)%data=-star%col(1)%data
         end if
          
         write(*,*) '* Now measuring shapes.'
         allocate(flux_s(nstar), flux_e_s(nstar))
         ! Now measure the flux with a tighter mask.
         ! Reset Gaussian width.
         a_par(1:2)=a_par(1:2)/2.0
         a_par(4)=1.0
         do i=1, nstar
           if (1000*(i/1000) == i) then
             print*, 'Done ', i, ' out of ', nstar, ' stars.'
           end if
           if (star(i)%col(1)%flg == 'OO') then
             noise=noise_map(nint(star(i)%x), nint(star(i)%y))
             ! Set position for this star.
             a_par(5)=star(i)%x
             a_par(6)=star(i)%y
             flux_s(i)=sum_flux(data, pix_flg, 0.0, 0.0, noise, &
             a_par, 2.0*seeing, 0.0, gain, low, high, flux_e_s(i), &
             cflag)
             if ('O'//cflag /= star(i)%col(1)%flg) then
               ! If the flags arn't the same, make sure a bad one is taken.
               if (star(i)%col(1)%flg=='OO') star(i)%col(1)%flg='O'//cflag
             end if
           end if
         end do         
         
         write(*,*) ' '
         
         if (nx>=0 .and. ny>=0) call halley(flux_s, flux_e_s, &
         star, nstar, nx, ny, ifname)
         
         deallocate(flux_s, flux_e_s)

       end subroutine shape_filter
      
      
       subroutine halley(flux_s, flux_s_e, star, nstar, nx, ny, ifname)

         real, dimension(:), intent(in) :: flux_s, flux_s_e
         !real, dimension(:), intent(in) :: apflux, apflux_e
         type(a_star), intent(inout), dimension(:) :: star
         integer, intent(in) :: nstar, nx, ny
         character(len=*) :: ifname

         integer :: n_flagged, n_tiny_err, n_sn, npts, istar, i, n_med
         real, dimension(nstar) :: z, sigma, x, y, resid, flux, flux_e
         real :: a, rms, chisq
         real, dimension(nx) :: bx
         real, dimension(ny) :: by
         real :: med_ratio, work
         integer :: iclip, iflag

         npts=0
         search: do istar=1, nstar
           if (star(istar)%col(1)%flg=='OO') then
             if (flux_s_e(istar) > tiny(flux_s_e(istar))) then
               if (star(istar)%col(1)%err > tiny(star(istar)%col(1)%err)) then
                 work=sqrt((flux_s_e(istar)/ & 
                 flux_s(istar))**2.0 + (star(istar)%col(1)%err/star(istar)%col(1)%data**2.0))
                 if (work < 0.01) then
                   npts=npts+1
                   z(npts) = flux_s(istar)/star(istar)%col(1)%data
                   sigma(npts) = work*(flux_s(istar)/star(istar)%col(1)%data)
                   x(npts) = star(istar)%x
                   y(npts) = star(istar)%y
                   flux(npts) = star(istar)%col(1)%data
                   flux_e(npts) = star(istar)%col(1)%err
                   if (npts == 1000) exit search
                 end if
               end if
             end if 
           end if
         end do search

         print*, 'Fitting ', npts, ' stellar shapes.'
         ! Add a systematic error.
         sigma(1:npts)=sqrt(sigma(1:npts)**2.0 + (0.01*z(1:npts))**2.0)
         call planefit_clip(z, sigma, x, y, npts, nx, ny, a, bx, by, rms, &
         chisq, iclip, resid, iflag)
         
         if (iflag == 0) then
           print*, 'Which gave parameters ', a, bx, by
         else
           print*, 'Fit failed with planefit_clip flag of ', iflag
           print*, 'No stellar shape rejection will be done.'
         end if
         
         open(unit=40, file=trim(ifname)//'.med', status='unknown')
         write(40,*) 'RMS of fit was ', rms, 'with a chi-squared of ', chisq
         write(40,*) 'After clipping ', iclip, ' points.'
         write(40,*) 'Columns are x, y, residual, error flux and flux error.'
         do istar=1, npts
           write(40,*) x(istar), y(istar), resid(istar), sigma(istar), &
           flux(istar), flux_e(istar)
           write(41,*) x(istar), y(istar), z(istar), sigma(istar), &
           flux(istar), flux_e(istar)
         end do
         close(40)
         
         n_flagged=0
         n_tiny_err=0
         n_sn=0
         n_med=0
         reject_loop: do istar=1, nstar
           if (star(istar)%col(1)%flg /= 'OO') then
             ! Flagged.
             n_flagged=n_flagged+1
             cycle reject_loop
           end if
           if (flux_s_e(istar) < tiny(flux_s_e(istar)) .or. &
             star(istar)%col(1)%err < tiny(star(istar)%col(1)%err)) then
             ! Avoid crashes in the next if statement.
             n_tiny_err=n_tiny_err+1
             cycle reject_loop
           end if
           if (flux_s(istar)/flux_s_e(istar)<3.0 .or. &
             ! And ones with signal-to-noise < 3
             star(istar)%col(1)%data/star(istar)%col(1)%err<3.0) then
             n_sn=n_sn+1
             cycle reject_loop
           end if
           if (iflag == 0) then
             ! What should the shape parameter be?
             med_ratio = a
             if (nx > 0) then
               do i=1, nx
                 med_ratio = med_ratio + bx(i)*(star(istar)%x**real(i))
               end do
             end if
             if (ny > 0) then
               do i=1, ny
                 med_ratio = med_ratio + by(i)*(star(istar)%y**real(i))
               end do
             end if
             ! Now work out the error on the ratio.
             work = (flux_s_e(istar)/flux_s(istar))**2.0
             work = work + (star(istar)%col(1)%err/star(istar)%col(1)%data)**2.0
             work = sqrt(work)*flux_s(istar)/star(istar)%col(1)%data
             ! Take 3 sigma, including RMS of fit.
             work=3.0*sqrt(work**2.0 + rms**2.0)
             if (abs((flux_s(istar)/star(istar)%col(1)%data)-med_ratio) > work) then
               n_med=n_med+1
               ! call write_star(28,star(istar),1)
               star(istar)%col(1)%flg='ON'
               cycle reject_loop
             end if
           end if         
         end do reject_loop
           
         print*, 'Number with zero error ', n_tiny_err
         print*, 'Number with S/N < 3 ', n_sn
         print*, 'Number rejected by median filter ', n_med

       end subroutine halley

     end module shape
