  module var_apcor_mod

    implicit none

    contains

    subroutine planefit_clip(z, sigma, x, y, npts, nx, ny, a, bx, by, rms, &
                             chisq, iclip, resid, iflag)

      real, dimension(:), intent(in) :: z, sigma, x, y
      integer, intent(in) :: npts, nx, ny
      real, intent(out) :: a
      real, dimension(:), intent(out) :: bx, by
      real, intent(out) :: rms, chisq
      integer, intent(out) :: iclip
      real, intent(out), dimension(:) :: resid
      ! Normally zero, -1 if too few data points.
      integer :: iflag
      
      integer :: jclip, kclip, lclip
      real, dimension(npts) :: weight
      real :: chi_next
      real, parameter :: sig_clip=3.0
      
      iflag=0

      weight(1:npts)=1.0/(sigma(1:npts)**2.0)

      clipping: do lclip=0, npts
        ! Clip out lclip points.
        if (lclip > 0) then
          do jclip=1, lclip
             call residual(z, x, y, npts, nx, ny, a, bx, by, resid)
             kclip=minval(maxloc(weight(1:npts)*resid(1:npts)**2.0 ) )
            weight(kclip)=0.0
          end do
        end if
        if (npts-lclip < 1+nx+ny) then
          ! Help we've run out of data to fit.
          a=0.0
          bx=0.0
          by=0.0
          rms=0.0
          iclip=0
          resid=0.0
          iflag=-1
          return
        end if
        call surfacefit(z, weight, x, y, npts, nx, ny, a, bx, by)
        call residual(z, x, y, npts, nx, ny, a, bx, by, resid)
        chisq=sum(weight(1:npts)*(resid(1:npts)**2.0))
        chisq=chisq/real(npts-lclip)
        ! Which would be the next point we'd clip?
        kclip=minval(maxloc(weight(1:npts)*(resid(1:npts)**2.0)))
        ! What's its chi-squared?
        chi_next = weight(kclip)*(resid(kclip)**2.0)
        ! Is it less than 3 sigma.
        if (chi_next < (sig_clip*2.0)*chisq) exit clipping
        ! No, then set all the weights back to normal.
        weight(1:npts)=1.0/(sigma(1:npts)**2.0)
      end do clipping

      jclip=0
      rms=0.0
      do lclip=1, npts
        if (weight(lclip) > 0.0) then
          rms = rms + resid(lclip)**2.0
          !(a + bx*x(lclip) + by*y(lclip) - z(lclip))**2.0 
          jclip=jclip+1
        end if
      end do
      rms=sqrt(rms/real(jclip))
      iclip=npts-jclip

    end subroutine planefit_clip      


!    subroutine surfacefit(z, weight, x, y, npts, nx, ny, c, bx, by)
!
!      ! Fits a surface to data z(x,y), where each point is weighted
!      ! by weight=1/sigma**2
!      ! This version uses a NAG routine.
!
!      real, dimension(:), intent(in) :: z, weight, x, y
!      integer, intent(in) :: npts, nx, ny
!      real, intent(out) :: c
!      real, dimension(:), intent(out) :: bx
!      real, dimension(:), intent(out) :: by
!
!      ! Local variables.
!      integer :: npars
!
!      integer :: ifail, i
!      integer, dimension(1+nx+ny) :: ipiv
!      
!      double precision, dimension(1+nx+ny) :: work1, work2, work3
!      double precision, dimension(npts) :: work4
!      double precision, dimension(npts, 1+nx+ny) :: qr, a
!      double precision, dimension(npts,1) :: b
!      double precision :: eps
!      double precision, dimension(1+nx+ny) :: alpha
!      double precision, dimension(1+nx+ny,1) :: pars
!      
!      ! External functions
!      double precision, external :: X02AJF
!
!      
!      a(1:npts,1)=dble(sqrt(weight(1:npts)))
!
!      if (nx > 0) then
!        do i=1,nx
!          a(1:npts,i+1)=dble(sqrt(weight(1:npts))*x(1:npts)**real(i))
!        end do
!      end if
!
!      if (ny > 0) then
!        do i=1,ny
!          a(1:npts,i+nx+1)=dble(sqrt(weight(1:npts))*y(1:npts)**real(i))
!        end do
!      end if
!
!      b(1:npts,1)=dble(sqrt(weight(1:npts))*z(1:npts))
!      
!      eps = X02AJF()	        
!      ifail = 0
!      npars = 1+nx+ny
!      
!      call F04AMF(a, npts, pars, npars, b, npts, npts, npars, 1, eps, qr, &
!      npts, alpha, work1, work2, work3, work4, ipiv, ifail)
!      
!
!      c=pars(1,1)
!      if (nx > 0) bx(1:nx)=pars(2:nx+1,1)
!      if (ny > 0) by(1:ny)=pars(2+nx:npars,1)
!
!    end subroutine surfacefit

    subroutine surfacefit(z, weight, x, y, npts, nx, ny, c, bx, by)

      ! Fits a surface to data z(x,y), where each point is weighted
      ! by weight=1/sigma**2
      ! This version calls a Starlink PDA routine.

      real, dimension(:), intent(in) :: z, weight, x, y
      integer, intent(in) :: npts, nx, ny
      real, intent(out) :: c
      real, dimension(:), intent(out) :: bx
      real, dimension(:), intent(out) :: by

      ! Locals.
      integer :: i
      integer :: ncols, mrows
      double precision, allocatable, dimension(:,:) :: w
      double precision, dimension(1+nx+ny) :: bl, bu
      integer, dimension(1+nx+ny) :: ind
      integer, dimension(1) :: iopt
      double precision, dimension(1+nx+ny) :: pars
      double precision :: rnorm
      integer :: mode
      double precision, dimension(5*(1+nx+ny)) :: rw
      integer, dimension(2*(1+nx+ny)) :: iw
      integer :: status


      ! Allocate w and fill it up.
      ncols=1+nx+ny
      mrows=npts
      allocate(w(mrows, ncols+1))
      
      w(1:npts,1)=dble(sqrt(weight(1:npts)))

      if (nx > 0) then
        do i=1,nx
          w(1:npts,i+1)=dble(sqrt(weight(1:npts))*x(1:npts)**real(i))
        end do
      end if

      if (ny > 0) then
        do i=1,ny
          w(1:npts,i+nx+1)=dble(sqrt(weight(1:npts))*y(1:npts)**real(i))
        end do
      end if

      w(1:npts,ncols+1)=dble(sqrt(weight(1:npts))*z(1:npts))

      status=0
      ind=4
      iopt(1)=99
      call pda_dbols(w, mrows, mrows, ncols, bl, bu, ind, iopt, pars, &
           rnorm, mode, rw, iw, status)

      if (mode < 0) then
        print*, 'Error reported by s/r pda_dbols.'
        print*, 'variable mode (which should be >= 0) is ', mode
        stop
      end if
      if (status /= 0) then
        print*, 'Error reported by s/r pda_dbols.'
        print*, 'variable status (which should be zero) is ', status
        stop
      end if

      c=pars(1)
      if (nx > 0) bx(1:nx)=pars(2:nx+1)
      if (ny > 0) by(1:ny)=pars(2+nx:ncols)

      deallocate(w)

    end subroutine surfacefit

    subroutine residual(z, x, y, npts, nx, ny, c, bx, by, resid)

      real, dimension(:), intent(in) :: z, x, y, bx, by
      integer, intent(in) :: npts, nx, ny
      real, intent(in) :: c
      real, dimension(npts), intent(out) :: resid
      integer :: i

      resid(1:npts) = z(1:npts) - c

      if (nx > 0) then
        do i=1, nx
          resid(1:npts) = resid(1:npts) - bx(i)*x(1:npts)**real(i)
        end do
      end if

      if (ny > 0) then
        do i=1, ny
          resid(1:npts) = resid(1:npts) - by(i)*y(1:npts)**real(i)
        end do
      end if


    end subroutine residual

    subroutine var_apcor(file_root, ifile, nfile, aprad, systematic, nx, ny, &
      bl_pfcor, nstar, iflag)
        
      ! Calculates the variation in the aperture correction as a
      ! a function of position.
      
      use ark_file_io
      use opt_extr
      use define_star
      
      ! The bit of the file name of the image before the dot.
      character(len=*), intent(in) :: file_root
      ! The current file and total number of files.
      integer :: ifile, nfile
      ! The radius of the aperture used in the aperture photmetry.
      real, intent(in) :: aprad
      real, intent(in) :: systematic
      ! The number of terms in each direction.
      integer, intent(in) :: nx, ny
      ! The corrections at the bottom left corner of the CCD.
      real, dimension(:,:), allocatable :: bl_pfcor
      ! Number of stars used for this ccd, and the total number for the file.
      integer, intent(out) :: nstar
      ! An error flag.  Zero for success < 0 to report errors.
      !     -1      .varin file not present.
      !     -4      No stars without fitting problems.
      !     -100+   a sizark error.
      !     -200+   an inpark error.
      !     -300+   an makark error.
      integer, intent(out) :: iflag
      
      ! Locals.
      ! The uncertainty above which we will not use a star in the profile 
      ! correction.
      real, parameter :: worst=0.03
      ! The constant term in the aperture correction.
      real :: a
      ! The change of aperture correction with position in x and y.
      real, allocatable, dimension(:) :: bx, by
      logical :: there
      character(len=8) :: arktyp
      ! Unit number for the .opt file.
      integer :: iunit_opt
      ! The log file and single star aperture corrections unit numbers.
      integer :: ilog, iap
      ! A general iostat.
      integer :: iostat
      ! The constant from the straight line fit (thrown away), and the rms.
      real :: rms, chisq
      ! The residuals.
      real, dimension(:), allocatable :: resid
      integer :: fit_flag
      integer :: jflag
      character :: cflag
      
      type(a_star), allocatable, dimension(:) :: star
      real, allocatable, dimension(:) :: apcor, apcorerr
      
      ! The sky, noise on the sky, and the error in the sky estimate.
      real :: sky, skynos, skyerr
      ! The sky box size.
      integer :: isky_box
      ! The input image.
      real, allocatable, dimension(:,:) :: data, axdata
      character, allocatable, dimension(:,:) :: pix_flg
      integer, dimension(2) :: naxis
      ! The available good area.
      integer, dimension(2) :: low, high
      ! Other parameters of CCD.
      real :: adu
      ! Star counters.
      integer :: istar, jstar, kstar, mstar, nstar_g
      ! The value and error for the aperture photometry.
      real :: apflux, aperr
      ! The number of points clipped.
      integer :: iclip
      real :: work
      integer :: i
      character(len=2) :: chext
      logical :: enough
      
      ! For the debugger.
      logical :: debug
      common / bug /debug
      
      call setbug()
      if (debug) call opt_debug_on()

      write(66,*) 'Doing file ', file_root
      
      ! Start as we mean to go on.
      iflag = 0

      allocate(bx(max(1,nx)))
      allocate(by(max(1,ny)))
      
      call arkinq(file_root, 'varin', there, arktyp)
      if (.not. there) then
        if (nx > 0) bx=0.0
        if (ny > 0) by=0.0
        mstar = 0
        iflag=-1
        return
      end if
      call arkopn(iunit_opt, file_root, 'varin', 'old', 'readonly', &
      'formatted', ' ', 0)
      read (iunit_opt,'(/,/)')
      ! Now count the number of stars.
      mstar=0
      star_count: do
        read(iunit_opt,*, iostat=iostat) work
        if (iostat /= 0) exit star_count
        mstar=mstar+1
      end do star_count
      ! Allocate some arrays.
      allocate(star(mstar), apcor(mstar), apcorerr(mstar))
      rewind(iunit_opt)
      ! Skip the header.
      read (iunit_opt,'(/,/)')

      kstar=0

      extension: do

        if (kstar == 0) then
          nstar=0
        else
          star(1)=star(nstar+1)
          nstar=1
        end if
        read_loop: do 
          kstar=kstar+1
          if (kstar > mstar) exit read_loop
          iostat=read_star(iunit_opt, star(nstar+1), 1)
          if (nstar > 1) then
            if (star(nstar)%ccd /= star(nstar+1)%ccd) exit read_loop
          end if
          nstar=nstar+1
        end do read_loop
        if (debug) print*, '@ Number of stars read was ', nstar
      
        ! Now read in the image.
        call nxtark_in(file_root, star(1)%ccd)
        call extark('fit')       
        iflag=inpark(naxis, data, axdata)
        if (iflag < 0) then
          iflag = iflag-200
          if (nx > 0) bx=0.0
          if (ny > 0) by=0.0
          nstar = 0
          return
        else
          iflag=0 
        end if
        jflag=inpflg(naxis, pix_flg)
      
        if ( & 
          get_header_i('ARK-LO1', low(1))  < 0 .or.   &
          get_header_i('ARK-LO2', low(2))  < 0 .or.   &
          get_header_i('ARK-HI1', high(1)) < 0 .or.  &
          get_header_i('ARK-HI2', high(2)) < 0) then
          print*, 'The active area needs setting for ', file_root
          print*, 'using the program pix_info.'
          stop
        end if
        if (get_header_r('ARK-GAIN', adu) < 0) then
          print*, 'The gain needs setting for ', file_root
          print*, 'using the program pix_info.'
          stop
        end if
      
        ! The sky box is the same size as the output image.  It is calculated 
        ! from the aperture size, in the way outlined in Naylor (1998).
        write(chext, '(i2.2)') star(1)%ccd
        call arkopn(iap, trim(file_root)//'_'//chext, 'apcors', 'unknown', &
        'overwrite', 'formatted', ' ', 0)
        write(iap,*) '! Columns are x position, y position, aperture'
        write(iap,*) '! correction, error in correction, uncorrected optimal magnitude'
        write(iap,*) '! and error in aperture magnitude.'
        isky_box=int (sqrt( 628.4*aprad*aprad))
        istar=1
        fit: do jstar=1, nstar
          print*, 'Fitting sky for ', jstar, '=', star(istar)%id
          call skyfit(data, star(istar)%x, star(istar)%y, &
          real(isky_box)/5.0, isky_box, low, high, &
          pix_flg, sky, skyerr, skynos, cflag)
          if (cflag == 'I') then
            write(*,*) 'Warning.  Star ', star(istar)%id, 'gave a bad fit to the sky.'
            !star(istar:nstar-1)=star(istar+1:nstar)
            !cycle fit
          else if (cflag /= 'O') then
            write(*,*) 'Rejecting star ', star(istar)%id, &
            '  as the sky fit failed with flag', cflag
            star(istar:nstar-1)=star(istar+1:nstar)
            cycle fit
          end if
          print*, 'Flux finding. Sky is ', sky
          apflux=sum_aper(data, aprad, sky, skyerr, &
          skynos, star(istar)%x, star(istar)%y, adu, &
          low, high, pix_flg, aperr, cflag)
          if (cflag/='O' .or. apflux<0.0) then
            write(*,*) 'Rejecting star ', star(istar)%id, &
            '  as the aperture photometry failed.'
            star(istar:nstar-1)=star(istar+1:nstar)
            cycle fit
          end if
          if (aperr > worst*apflux) then
            ! We don't want points with a large uncertainty.  We can then
            ! filter out solutions that have a small number of poor points.
            write(*,*) 'Rejecting star ', star(istar)%id, &
            '  as the signal-to-noise is worse than ', 100.0*worst, ' percent.'
            star(istar:nstar-1)=star(istar+1:nstar)
            cycle fit
          end if
          ! Convert into magnitudes.
          aperr=aperr/apflux
          apflux=-2.5*log10(apflux)
          apcor(istar) = apflux - star(istar)%col(1)%data
          apcorerr(istar) = sqrt(aperr**2.0 + star(istar)%col(1)%err**2.0) 
          write(iap,*) star(istar)%x, star(istar)%y, apcor(istar), &
          apcorerr(istar), star(istar)%col(1)%data, star(istar)%col(1)%err 
          istar=istar+1
        end do fit
        close(iap)
        deallocate(data, axdata, pix_flg)

        nstar_g=istar-1

        enough=.true.
        if (nstar_g == 0) then
          enough=.false.
          iflag = -4
          if (nx > 0) by = 0.0
          if (ny > 0) bx = 0.0

        else

          ! O.K., so how good do we think the profile correction is?
          work=sqrt(1.0/sum(1.0/(apcorerr(1:nstar_g)**2.0 + systematic**2.0)))
          if (work > 0.3*systematic) then
            ! If we use 0.01 in the above statement it means for a straight 
            ! line the uncertainty in the profile correction would be 1 
            ! percent.
            enough=.false.
          end if

          if (enough) then
            write(66,*) 'Got a noise value of ', work
          else
            write(66,*) 'Got a noise value of ', work, ' which will not give a good correction.'
          end if

          allocate(resid(nstar_g))

          print*, 'Fitting corections.'

          apcorerr(1:nstar_g) =  sqrt(apcorerr(1:nstar_g)**2.0 + systematic**2.0)
          call planefit_clip(apcor, apcorerr, star%x, star%y, nstar_g, &
          nx, ny, a, bx, by, rms, chisq, iclip, resid, fit_flag)

          bl_pfcor(star(1)%ccd, ifile)=a

          ! Open the log file.
          call arkopn(ilog, trim(file_root)//'_'//chext, 'varlog', 'unknown', &
          'overwrite', 'formatted', ' ', 0)
          write(ilog,*) '! RMS of fit is ', rms, ' after clipping ', iclip, &
          ' points.'
          work=a
          if (nx > 0) then
            do i=1, nx
              work=work+bx(i)*((real(naxis(1))/2.0)**real(i))
            end do
          end if
          if (ny > 0) then
            do i=1, ny
              work=work+by(i)*((real(naxis(2))/2.0)**real(i))
            end do
          end if
          write(ilog,*) '! Single star aperture correction at field center is ', &
          work
          write(ilog,*) '! Columns are x, y, residual and error.  Chi**2 is ', &
          chisq
          do istar=1, nstar_g
            write(ilog,*) star(istar)%x, star(istar)%y, resid(istar), apcorerr(istar)
          end do
          close(ilog)

          deallocate(resid)

        end if

        ! Write out the new co-efficients
        write(1,*) file_root, star(1)%ccd
        if (nx>0 .and. ny>0) then
          write(1,*) enough, a, bx(:), by(:)
        else if (nx > 0) then
          write(1,*) enough, a, bx(:)
        else if (ny > 0) then
          write(1,*) enough, a, by(:)
        else
          write(1,*) enough, a
        end if

        if (kstar > mstar) exit extension

      end do extension
      close(iunit_opt)
      deallocate(bx)
      deallocate(by)

    end subroutine var_apcor

    subroutine var_pfcor(file_root, systematic, nx, ny, bl_pfcor)

      use define_star
      use ark_file_io

      character(len=*), intent(in) :: file_root
      real, intent(in) :: systematic
      ! The number of terms in each direction.
      integer, intent(in) :: nx, ny
      ! The corrections at the bottom left corner of the CCD.
      real, dimension(:) :: bl_pfcor

      character(len=2) :: chext
      integer :: ipf, ilog, iostat
      type(a_star), dimension(:), allocatable :: star
      real, dimension(:), allocatable :: pfcor, pfcorerr, resid
      integer :: nstar, istar
      ! The constant term in the aperure correction.
      real :: a
      ! The change of aperture correction with position in x and y.
      real, allocatable, dimension(:) :: bx, by
      real :: rms, chisq, work
      integer :: iclip
      integer :: fit_flag
      logical :: there, enough
      integer :: iext

      allocate(bx(max(1,nx)))
      allocate(by(max(1,ny)))
          
      extension: do iext=1, 99

        write(chext, '(i2.2)') iext
        inquire(file=trim(file_root)//'_'//chext//'.apcors', exist=there)
        if (.not. there) cycle extension
        call arkopn(ipf, trim(file_root)//'_'//chext, 'apcors', 'old', &
        'readonly', 'formatted', ' ', 0)
        read(ipf,'(/,/)')
        allocate(star(1), pfcor(1), pfcorerr(1))
        nstar=0
        count_them: do
          read(ipf,*,iostat=iostat) star(1)%x, star(1)%y, pfcor(1), &
          pfcorerr(1), star(1)%col(1)%data, star(1)%col(1)%err 
          if (iostat < 0) exit count_them
          nstar=nstar+1
        end do count_them

        enough=.true.

        if (nstar == 0) then

          enough=.false.
          if (nx > 0) by = 0.0
          if (ny > 0) bx = 0.0
          close(ipf)

        else

          rewind(ipf)
          read(ipf,'(/,/)')
          deallocate(star, pfcor, pfcorerr)
          allocate(star(nstar), pfcor(nstar), pfcorerr(nstar))
          do istar=1, nstar
            read(ipf,*) star(istar)%x, star(istar)%y, pfcor(istar), &
            pfcorerr(istar), star(istar)%col(1)%data, star(istar)%col(1)%err 
          end do
          close(ipf)

          ! O.K., so how good do we think the profile correction is?
          work=sqrt(1.0/sum(1.0/(pfcorerr(1:nstar)**2.0 + systematic**2.0)))
          if (work > 0.3*systematic) then
            ! If we use 0.01 in the above statement it means for a straight 
            ! line the uncertainty in the profile correction would be 1 
            ! percent.
            enough=.false.
          end if

          allocate(resid(nstar))

          pfcorerr(1:nstar) =  sqrt(pfcorerr(1:nstar)**2.0 + systematic**2.0)
          call planefit_clip(pfcor, pfcorerr, star%x, star%y, nstar, &
          nx, ny, a, bx, by, rms, chisq, iclip, resid, fit_flag)

          ! Write out the new co-efficients
          write(1,*) file_root, iext
          if (nx>0 .and. ny>0) then
            write(1,*) enough, a, bx(:), by(:)
          else if (nx > 0) then
            write(1,*) enough, a, bx(:)
          else if (ny > 0) then
            write(1,*) enough, a, by(:)
          else
            write(1,*) enough, a
          end if

          bl_pfcor(iext)=a

          ! Open the log file.
          call arkopn(ilog, trim(file_root)//'_'//chext, 'varlog', 'unknown', &
          'overwrite', 'formatted', ' ', 0)
          write(ilog,*) '! RMS of fit is ', rms, ' after clipping ', iclip, &
          ' points.'
          write(ilog,*) '! Single star profile correction at (1,1) is ', a
          write(ilog,*) '! Columns are x, y, residual and error.  Chi**2 is ', &
          chisq
          do istar=1, nstar
            write(ilog,*) star(istar)%x, star(istar)%y, resid(istar), pfcorerr(istar)
          end do
          close(ilog)

          deallocate(resid, star, pfcor, pfcorerr)

        end if

      end do extension
      deallocate(bx)
      deallocate(by)

    end subroutine var_pfcor




    
        subroutine clean_line(header)

        character(len=*), intent(inout) :: header

        integer :: i

        clean: do i=1, len(header)
          if (header(i:i) == '.') cycle clean
          if (header(i:i) == '-') cycle clean
          if (ichar(header(i:i))>47 .and. ichar(header(i:i))<58) cycle clean
          header(i:i)=' '
        end do clean

        end subroutine clean_line

  end module var_apcor_mod
