    program bradley

      use ark_file_io
      use opt_extr
      use shape

      implicit none

      character(len=80) :: root, root_and_ext

      ! The profile shape from the PSF star.
      real, dimension(6) :: a_par
      ! The clipping radius for the optimal photometry.
      real :: cliprad
      real :: skycnt, skyerr, skynos
      integer :: ibox
      real :: fwhm
      character :: cflag, sky_flag

      character(len=10) :: chdummy

      type(a_star), allocatable, dimension(:) :: star
      real, allocatable, dimension(:) :: flux_s, flux_e_s
      type(a_colour), allocatable, dimension(:) :: save_col

      integer :: ncol
      character(len=10), dimension(mcol) :: colstr

      ! For the image.
      integer, dimension(2) :: naxis, low, high
      integer :: iflag, iext
      real, dimension(:,:), allocatable :: data, axdata
      character, dimension(:,:), allocatable :: pix_flg
      real :: gain

      integer :: nx, ny

      ! The number of stars, and counters.
      integer :: nstar, istar, jstar, istart, iend

      character(len=100) :: header
      integer :: iostat
      integer, parameter :: infil=10
      integer :: iunit
      logical :: first

      logical :: ap_phot
      real :: aprad

      call setbug()

      print*, '> Give orders for polynomial.'
      read(*,*) nx, ny
      first=.true.

      frame: do
        call extark('fit')
        ! Read in the data.  The image is contained in the array data.
        iflag=inpark(naxis, data, axdata)
        if (iflag < 0) exit frame
        ! And any flags.
        iflag=inpflg(naxis, pix_flg)
        ! Recover the name of the last file read by lstark.
        call lstark_in(root, iext)
        ! Remove leading blanks.
        root=adjustl(root)
        iflag=index(root, '.')
        if (iflag /= 0) root(iflag:len(root))=' '

        if (iext == 0) then
          ! First time around so get all the data.
          print*, 'Reading ', trim(root)//'_opt.cat'
          call nxtcls_in(trim(root)//'_opt.cat')
          iostat=read_cluster_file(nstar, ncol, colstr, star)
          if (iostat /= 0) then
            print*, 'Error reading ', trim(root)//'_opt.cat'
            stop
          end if
          ! Decode the header.  First is this aperture or optimal photometry?
          read (cluster_header(1),*) chdummy
          if (chdummy == 'Aperture') then
            ap_phot=.true.
            read (cluster_header(1),*) chdummy, chdummy, aprad
            print*, 'Writing it out to ', trim(root)//'_old_ap.cat'
            call nxtcls_out(trim(root)//'_old_ap.cat')
          else
            ! Read first the seeing, and then the rotation, and 
            ! finally the clipping radius.
            ! I do this by removing everything but the numbers, and 
            ! reading those.
            ap_phot=.false.
            read (cluster_header(1),*) chdummy, chdummy, a_par(1:3), cliprad
            fwhm=sqrt(a_par(1)*a_par(2))
            a_par(1)=a_par(1)/1.665
            print*, 'Writing it out to ', trim(root)//'_old_opt.cat'
            call nxtcls_out(trim(root)//'_old_opt.cat')
          end if
          ! Now, write it out somewhere safe.
          call start_cluster_file(ncol, colstr, iunit, nstar)
          do istar=1, nstar
            call write_star(iunit, star(istar), ncol)
          end do 
          close(iunit)
          ! Before we start faffing with them, save the current "colour".
          allocate(save_col(nstar))
          save_col=star%col(1)
          ! We don't have to allocate these arrays this large (they only need
          ! be the size of the extension), but doing so make the code
          ! simpler.
          allocate(flux_s(nstar), flux_e_s(nstar))
        end if

        if (naxis(1)*naxis(2) > 0) then
          if ( & 
            get_header_i('ARK-LO1', low(1))  < 0 .or.   &
            get_header_i('ARK-LO2', low(2))  < 0 .or.   &
            get_header_i('ARK-HI1', high(1)) < 0 .or.  &
            get_header_i('ARK-HI2', high(2)) < 0) then
            print*, 'The active area needs setting for ', root
            print*, 'using the program pix_info.  Going to next file.'
            deallocate(data, axdata, pix_flg)
            cycle frame
          end if
          if (get_header_r('ARK-GAIN', gain) < 0) then
            print*, 'The gain needs setting for ', root
            print*, 'using the program pix_info.  Going to next file.'
            deallocate(data, axdata, pix_flg)
            cycle frame
          end if

          ! Find out where this extension starts and ends.
          find_start: do istar=1, nstar
            if (star(istar)%ccd == iext) then
              istart=istar
              exit find_start
            end if
          end do find_start
          iend=nstar
          find_end: do istar=istart+1, nstar
            if (star(istar)%ccd /= iext) then
              iend=istar-1
              exit find_end
            end if
          end do find_end

          print*, 'Extension ', iext, ' starts at star ', istart, ' and ends at star ', iend
          print*, 'and there are ', iend-istart+1, ' stars for this extension.'

          write(*,*) '* Now measuring shapes.'
          if (.not. ap_phot) then
            ! Now measure the flux with a tighter mask.
            ! Reset Gaussian width.
            a_par(1:2)=a_par(1:2)/2.0
            a_par(4)=1.0
          else
            aprad=aprad/2.0
            fwhm=0.0
          end if
          jstar=0
          each_star: do istar=istart, iend
            if (star(istar)%col(1)%flg /= 'OO') cycle each_star
            jstar=jstar+1
            if (1000*(jstar/1000) == jstar) then
              print*, 'Done ', jstar, ' stars.'
            end if
            if (ap_phot) then
              ibox=int(sqrt(628.4*aprad*aprad))
            else
              ibox=int(sqrt(453.2*fwhm*fwhm+ 4.0*fwhm*fwhm))
            end if
            call skyfit(data, star(istar)%x, star(istar)%y, &
            fwhm, ibox, low, high, pix_flg, skycnt, skyerr, skynos, sky_flag)
            if (ap_phot) then
              flux_s(istar)=sum_aper(data, aprad, skycnt, skyerr, skynos, &
              star(istar)%x, star(istar)%y, gain, low, high, pix_flg, &
              flux_e_s(istar), cflag)
            else
              ! Set position for this star.
              a_par(5)=star(istar)%x
              a_par(6)=star(istar)%y
              flux_s(istar)=sum_flux(data, pix_flg, skycnt, skyerr, skynos, &
              a_par, 2.0*fwhm, 0.0, gain, low, high, flux_e_s(istar), cflag)
            end if
            if (sky_flag/='O' .and. cflag=='O') cflag=sky_flag
            if ('O'//cflag /= star(istar)%col(1)%flg) then
              ! If the flags arn't the same, make sure a bad one is taken.
              if (star(istar)%col(1)%flg=='OO') &
              star(istar)%col(1)%flg='O'//cflag
            end if
            call mag_to_flux(star(istar)%col(1))
            !print*, star(istar)%col(1), flux_s(istar), flux_e_s(istar)
          end do each_star

          ! Now add the enxtension number onto root.
          root_and_ext=root
          write(root_and_ext(len_trim(root)+1:len_trim(root)+3), '(a, i2.2)') &
          '_', iext
      
          if (nx>=0 .and. ny>=0) call halley(flux_s(istart:iend), &
          flux_e_s(istart:iend), star(istart:iend), iend-istart+1, nx, ny, root_and_ext)

          deallocate(data, axdata, pix_flg)

        end if

        if (iend == nstar) then
          ! Put the old flags and mags back, except for stuff that's 
          ! gained a non-stellar flag.
          where(star%col(1)%flg /= 'ON') 
            star%col(1)=save_col
          elsewhere
            star%col(1)%data=save_col%data
            star%col(1)%err =save_col%err
          endwhere
          deallocate(save_col)
        
          ! Write out to the new opt file.
          call nxtcls_out(trim(root)//'_opt.cat')
          call start_cluster_file(ncol, colstr, iunit, nstar)
          do istar=1, nstar
            call write_star(iunit, star(istar), 1)
          end do
          deallocate(star)
          close(iunit)
          deallocate(flux_s, flux_e_s)
        end if

      end do frame


    end program bradley

         
