  module detector                

    ! Release 2.0; by Tim Naylor

    use get_header

    implicit none

    integer, private, dimension(2) :: low_safe=(/1,1/), high_safe=(/-1,-1/)
    real, private :: gain_safe=1.0, read_safe=0.0, satur_safe=65537.0

    contains

      subroutine rd_detector(prlow, prhigh, prgain, prread, prsatur, &
      low, high, gain, read, satur)

!     Returns the properties of the detector from the detector.prm file.
!     If pr.... is set, and it doesn't find the value it prompt for it.  
!     If pr.... is set, and it does find the value, it prints it out.

!     It writes a detector.prm file from the inputs (if one isn't there),
!     or adds new properties to the current one.

!     Inputs.
!     If any property is not found, should the routine prompt for it?
      logical :: prlow, prhigh, prgain, prread, prsatur

!     Inputs/Outputs
!     The properties.  The input values are used as defaults, and those for
!     low and high should be set to 1 and naxis on input.
      integer, dimension(2) :: low, high
      real :: gain, read, satur

!     Locals
      logical :: there
      integer :: iunit, iostat, i
      character(len=50) :: input
      character(len=20) :: key
      logical :: nolow, nohigh, noread, nogain, nosatur
      integer, dimension(2) :: def_high

!     Save the default image size.
      def_high=high

      nolow=prlow
      nohigh=prhigh
      nogain=prgain
      noread=prread
      nosatur=prsatur      
      inquire(file='detector.prm', exist=there)
      if (there) then
!       Use an ARK routine to open the file in a machine indepenent manner.
!       In this case you are returned a unit number (iunit), which is an
!       already extant ASCII file called detector.prm.
        call arkopn(iunit, 'detector', 'prm', 'old', 'overwrite', &
        'formatted', ' ', 0)
        lines: do
          read(iunit, '(a)', iostat=iostat) input
!         Check for end of file.
          if (iostat < 0 ) exit lines
!         Find the equals sign
          search: do i=len(input), 1, -1
            if (input(i:i) .eq. '=') then
              read(input(1:i-1), '(a)') key
              key=adjustl(key)
              iostat=0
              if (key.eq.'area' .or. key.eq.'AREA') then
                read(input((i+1):len(input)),*,iostat=iostat) &
                low(1), high(1), low(2), high(2)
                if (iostat == 0) nolow=.false.
                if (iostat == 0) nohigh=.false.
              else if (key.eq.'gain' .or. key.eq.'GAIN') then
                read(input((i+1):len(input)),*,iostat=iostat) gain
                if (iostat == 0) nogain=.false.
              else if (key.eq.'read' .or. key.eq.'READ') then
                read(input((i+1):len(input)),*,iostat=iostat) read
                if (iostat == 0) noread=.false.
              else if (key.eq.'saturation' .or. key.eq.'SATURATION') then
                read(input((i+1):len(input)),*,iostat=iostat) satur
                if (iostat == 0) nosatur=.false.
              else 
                print*, '* WARNING, the keyword ', input(1:i-1)
                print*, '* of the detector.prm file was not understood.'
              end if
              if (iostat /= 0) then
                print*, '* WARNING, failed to read ', key
                print*, '* from detector.prm file.'
              end if
              cycle lines
            end if
          end do search
          print*, '* WARNING, line ', input
          print*, '* of the detector.prm file was not understood.'
        end do lines
      else
        print*, ' '
        print*, '* Creating a detector.prm file.'
!       Use an ARK routine to create a file in a machine indepenent manner.
!       The file is on logical unit iunit, and is called detector.prm.
        call arkopn(iunit, 'detector', 'prm', 'new', 'overwrite', &
        'formatted', ' ', 0)      
      end if

      if (nolow) then                 
        asklow1: do 
          print*, '> Lowest X pixel value to be used, <CR> = ', low(1)
          input=' '
          read (*,'(a)') input
          iostat=0
          if (input .ne. ' ') read (input,*,iostat=iostat) low(1)
          if (iostat == 0) exit asklow1
        end do asklow1
        asklow2: do 
          print*, '> Lowest Y pixel value to be used, <CR> = ', low(2)
          input=' '
          read (*,'(a)') input
          iostat=0
          if (input .ne. ' ') read (input,*,iostat=iostat) low(2)
          if (iostat == 0) exit asklow2
        end do asklow2
      else if (prlow) then
        print*, '* Lowest X pixel to be used is', low(1)
        print*, '* Lowest Y pixel to be used is', low(2)
      end if

      if (nohigh) then
        askhigh1: do 
          print*, '> Highest X pixel value to be used, <CR> = ', high(1)
          input=' '
          read (*,'(a)') input
          iostat=0
          if (input .ne. ' ') read (input,*,iostat=iostat) high(1)
          if (iostat == 0) exit askhigh1
        end do askhigh1
        askhigh2: do 
          print*, '> Highest Y pixel value to be used, <CR> = ', high(2)
          input=' '
          read (*,'(a)') input
          iostat=0
          if (input .ne. ' ') read (input,*,iostat=iostat) high(2)
          if (iostat == 0) exit askhigh2
        end do askhigh2
      else if (prhigh) then
        print*, '* Highest X pixel to be used is', high(1)
        print*, '* Highest Y pixel to be used is', high(2)
      end if

!     Check that the area found is reasonable.
      if (low(1) < 1) print*, 'WARNING lowest good X pixel is < 0!'
      if (low(2) < 1) print*, 'WARNING lowest good Y pixel is < 0!'
      if (high(1) > def_high(1)) & 
      print*, 'WARNING highest good X pixel is beyond image edge!'
      if (high(2) > def_high(2)) & 
      print*, 'WARNING highest good Y pixel is beyond image edge!'
      if ( 2*(high(1)-low(1)+1)*(high(2)-low(2)+1) < def_high(1)*def_high(2) ) &
      print*, 'WARNING less than half image is good pixels!'

      if (nolow .and. nohigh) then
        print*, '* Adding AREA keyword to detector.prm'
        write(iunit,100) low(1), high(1), low(2), high(2)
100     format(' area = ', i5, 1x, i5, 1x, i5, 1x, i5)
      end if

      if (nogain) then
        askgain: do 
          print*, '> Give number of photons per ADU, <CR> = ', gain
          input=' '
          read (*,'(a)') input
          iostat=0
          if (input .ne. ' ') read (input,*,iostat=iostat) gain
          if (iostat == 0) then
            print*, '* Adding GAIN keyword to detector.prm'
            write(iunit,*) 'gain = ', gain
            exit askgain
          end if
        end do askgain
      else if (prgain) then
        print*, '* Using a gain of ', gain, ' detected photons per count.'
      end if

      if (noread) then
        askread: do 
          print*, '> Give read noise in counts, <CR> = ', read
          input=' '
          read (*,'(a)') input
          iostat=0
          if (input .ne. ' ') read (input,*,iostat=iostat) read
          if (iostat == 0) then   
            print*, '* Adding READ keyword to detector.prm'
            write(iunit,*) 'read =', read
            exit askread
          end if
        end do askread
      else if (prread) then
        print*, '* Using a read noise of ', read, ' counts.'
      end if

      if (nosatur) then
        asksatur: do 
          print*, '> Give saturation in level in counts, <CR> = ', satur
          input=' '
          read (*,'(a)') input
          iostat=0
          if (input .ne. ' ') read (input,*,iostat=iostat) satur
          if (iostat == 0) then
            print*, '* Adding SATURATION keyword to detector.prm'
            write(iunit,*) 'saturation =', satur
            exit asksatur
          end if
        end do asksatur
      else if (prsatur) then
        print*, '* Using a saturation level of ', satur, ' counts.'
      end if

      close(iunit)
      print*, ' '

      ! Save the default parameters for later images.
      low_safe=low 
      high_safe=high
      gain_safe=gain
      read_safe=read
      satur_safe=satur

      end subroutine rd_detector

      subroutine up_detector(low, high, gain, read, satur)

      integer, dimension(2), intent(out) :: low, high
      real, intent(out) :: gain, read, satur

      real :: r_work
      integer :: i_work

      low=low_safe
      high=high_safe
      gain=gain_safe
      read=read_safe
      satur=satur_safe

      if (get_header_r('ARK-SAT', r_work) > 0) then
        satur=r_work
        print*, '* Saturation level for this image is ', satur, ' counts.'
      end if
      
      if (get_header_r('ARK-GAIN', r_work) > 0) then
        gain=r_work
        print*, '* Gain for this image is ', gain, ' electrons/count.'
      end if

      if (get_header_i('ARK-LO1', i_work) > 0) then
        low(1)=i_work
        print*, '* lowest X pixel for this image is ', low(1)
      end if
      if (get_header_i('ARK-LO2', i_work) > 0) then
        low(2)=i_work
        print*, '* lowest Y pixel for this image is ', low(2)
      end if
      
      if (get_header_i('ARK-HI1', i_work) > 0) then
        high(1)=i_work
        print*, '* highest X pixel for this image is ', high(1)
      else if (get_header_i('NAXIS1', i_work) > 0) then
        if (i_work < high(1)) then
          high(1)=i_work
          print*, '* Highest X pixel for this image is ', high(1)
        end if
      end if

      if (get_header_i('ARK-HI2', i_work) > 0) then
        high(2)=i_work
        print*, '* highest Y pixel for this image is ', high(2)
      else if (get_header_i('NAXIS2', i_work) > 0) then
        if (i_work < high(2)) then
          high(2)=i_work
          print*, '* Highest Y pixel for this image is ', high(2)
        end if
      end if

      end subroutine up_detector

  end module detector
