     program flamsteed

       use ark_file_io
       use flamsteed_subs
       use shape

       implicit none

       ! For reading in the data.
       real, allocatable, dimension(:,:) :: data, axdata
       integer, dimension(2) :: naxis
       integer :: jflag, iflag
       character(len=50) :: ifname, ofname_root, offset, file

       ! The detector characteristics.
       integer, dimension(2) :: low, high
       real :: gain, good_max
       ! The seeing and search sigma.
       real :: seeing, sigma
       ! The background box size.
       integer :: ibox
       ! The order of the PSF polynmials
       integer :: nx, ny

       ! For the significant events.
       integer, allocatable, dimension(:) :: x_sig, y_sig
       integer :: n_sig, i_sig

       ! For the output catalogue.
       real, dimension(:), allocatable :: ellip
       type(a_star), dimension(:), allocatable :: star, starl, star_all
       integer :: nstar, nstar_all, i, j, istar, iaxis1, iaxis2

       logical, dimension(:,:), allocatable :: bin_sig_map
       real, dimension(:,:), allocatable :: noise_map

       integer :: iext, jext, start_ccd, end_ccd
       logical :: there
       integer :: iunit1, iostat
       logical :: opened_offsets=.false.
       real, dimension(6) :: coff
       
       ! Good pixel and staurated pixel arrays.
       character, allocatable, dimension(:,:) :: pix_flg
       
       ! For the ellipticity.
       ! Second moments.
       real :: xx, yy, xy, inten
       real :: median_noise

       call setbug()

       print*, '* The seeing estimate is used both as the box size for'
       print*, '* sigma searching, and half the radius within which we'
       print*, '* believe a star is a duplicate.  So its safer to use'
       print*, '* a large rather than a small number.'
       print*, ' '

       print*, '> Give approximate seeing, and sigma for search.'
       read(*,*) seeing, sigma
       print*, ' '
       print*, '  To detect non-stellar objects, this program assumes the'
       print*, '  shape parameter is a polynomial function of position.'
       print*, &
         '> Give the X and Y order to be used (<0 for no discrimination).'
       read(*,*) nx, ny

       print*, 'First let''s deal with the long exposure.'

       extension_long: do
         iflag=inpark(naxis, data, axdata)
         if (iflag < 0) exit extension_long
         call lstark_in(ifname, iext)
         if (naxis(1)*naxis(2) == 0) cycle extension_long
         if ( & 
           get_header_i('ARK-LO1', low(1))  < 0 .or.   &
           get_header_i('ARK-LO2', low(2))  < 0 .or.   &
           get_header_i('ARK-HI1', high(1)) < 0 .or.  &
           get_header_i('ARK-HI2', high(2)) < 0) then
           print*, 'The active area needs setting for ', ifname
           print*, 'using the program pix_info.'
           deallocate(data,axdata)
           stop
         end if
         if (get_header_r('ARK-GAIN', gain) < 0) then
           print*, 'The gain needs setting for ', ifname
           print*, 'using the program pix_info.'
           deallocate(data,axdata)
           stop
         end if
         jflag=inpflg(naxis, pix_flg)

         ! The sky box size is determined in the way described in the paper
         ibox=int(sqrt(453.2*seeing*seeing+4.0*seeing*seeing))

         print*, 'Determining sky and noise levels.'
         call back(data, low, high, ibox, noise_map, median_noise, &
         pix_flg)
         print*, 'Median noise is ', median_noise

         nstar=naxis(1)*naxis(2)/(4*seeing*seeing)
         allocate(star(nstar))
         call zero_star(star)
         star%ccd=iext
         nstar=0
       
         allocate(bin_sig_map(size(data,1),size(data,2)))
         
         good_max=minval(data)
         do iaxis2=low(2), high(2)
           do iaxis1=low(1), high(1)
             if (pix_flg(iaxis1,iaxis2) == 'O') then
               good_max=max(good_max,data(iaxis1,iaxis2))
             end if
           end do
         end do
         
         do i=int(log(good_max/(sigma*median_noise))/log(1.4142)), 1, -1

           print*, 'Pass ', i, ' at ', real(1.4142**(i-1))*sigma, ' sigma.'
       
           call sig(data, low, high, real(1.4142**(i-1))*sigma, seeing, &
           noise_map, pix_flg, bin_sig_map)

           ! Add up the number of events.
           n_sig=count(bin_sig_map)
           allocate(x_sig(n_sig), y_sig(n_sig))
           i_sig=0
           do iaxis2=low(2)+1, high(2)-1
             do iaxis1=low(1)+1, high(1)-1
               if (bin_sig_map(iaxis1,iaxis2)) then
                 i_sig=i_sig+1
                 x_sig(i_sig)=iaxis1
                 y_sig(i_sig)=iaxis2
               end if
             end do
           end do
           print*, 'Found ', n_sig, 'events.'

           ! Now filter the events.
           call filter(x_sig, y_sig, n_sig, high, data, &
           pix_flg, star, nstar)
           print*, 'Which gives ', nstar, ' stars.'
           deallocate(x_sig, y_sig)
         end do
         
         deallocate(bin_sig_map)

         ! call doubles(seeing, star, nstar)
         
         ! print*, 'After removing duplicates, number of stars left was ', nstar

         i=index(ifname, '.') 
         if (i <= 0) i=len_trim(ifname)
         ofname_root=ifname(1:i-1)//'_'
         write(ofname_root(i+1:i+2),'(i2.2)') iext
         call shape_filter(data, star, nstar, seeing, low, &
         high, noise_map, gain, pix_flg, nx, ny, ofname_root, .true.)

         allocate(starl(nstar))
         ! Actually remove junk from the list.
         istar=0
         junkit: do i=1, nstar
           ! Stars too close to edge or outside region.
           if (star(i)%col(1)%flg=='OE') cycle junkit
           if (star(i)%col(1)%flg=='OR') cycle junkit
           istar=istar+1
           starl(istar)=star(i)
         end do junkit
         nstar=istar
         
         deallocate(star)         
         deallocate(noise_map)         

         print*, 'After removing low s:n and stars close to edge, number left was ', nstar
       
         call doubles(seeing, starl, nstar)

         print*, 'After removing duplicates again, number of stars left was ', nstar
         
         ! Find the ellipticity.
         ! open(unit=12, file=trim(ofname_root)//'.ell', status='unknown')
         ! write(12,*) 'Columns are star id number, ellipticity and S:N.'
         allocate(ellip(nstar))
         ell: do istar=1, nstar
           inten=0.0
           xx=0.0
           yy=0.0
           xy=0.0
           do i=nint(starl(istar)%x)-nint(2.0*seeing), nint(starl(istar)%x)+nint(2.0*seeing)
             if (i<low(1) .or. i>high(1)) cycle ell
             do j=nint(starl(istar)%y)-nint(2.0*seeing), nint(starl(istar)%y)+nint(2.0*seeing)
               if (j<low(2) .or. j>high(2)) cycle ell
               xx = xx + data(i,j)*((real(i)-starl(istar)%x)**2.0)
               yy = yy + data(i,j)*((real(j)-starl(istar)%y)**2.0)
               xy = xy + data(i,j)* (real(i)-starl(istar)%x)*&
               (real(j)-starl(istar)%y)
               inten=inten+data(i,j)
             end do                
           end do
           if (abs(xx+yy) > 2.0*tiny(xx+yy)) then
             ellip(istar)=sqrt( (xx-yy)**2.0 + 4.0*(xy**2.0))/(xx+yy)
             if (starl(istar)%col(1)%flg=='OO') &
             write(12,*) starl(istar)%id, ellip(istar), &
             starl(istar)%col(1)%data/starl(istar)%col(1)%err
             if (ellip(istar)<0.0) starl(istar)%col(1)%flg='ON'
           end if
         end do ell
         ! close(12)
         
         deallocate(data, axdata, pix_flg, ellip)

         if (.not. allocated(star_all)) then
           allocate(star_all(30*nstar))
           star_all(1:nstar)=starl(1:nstar)
           nstar_all=nstar
         else
           if (nstar_all+nstar > size(star_all)) then
             print*, nstar_all+nstar, size(star_all)
             stop
           end if
           star_all(nstar_all+1:nstar_all+nstar)=starl(1:nstar)
           nstar_all=nstar_all+nstar
         end if

         deallocate(starl)

       end do extension_long

       print*, 'Now let''s do the short exposure.'
       print*, 'Starting with ', nstar, ' stars.'
       print*, 'Type end if you don''t have a short exposure.'

       extension: do
         iflag=inpark(naxis, data, axdata)
         if (iflag < 0) exit extension
         if (.not. allocated(star)) allocate(star(nstar_all))
         call lstark_in(ifname, iext)
         if (naxis(1)*naxis(2) == 0) cycle extension
         if ( & 
           get_header_i('ARK-LO1', low(1))  < 0 .or.   &
           get_header_i('ARK-LO2', low(2))  < 0 .or.   &
           get_header_i('ARK-HI1', high(1)) < 0 .or.  &
           get_header_i('ARK-HI2', high(2)) < 0) then
           print*, 'The active area needs setting for ', ifname
           print*, 'using the program pix_info.'
           deallocate(data,axdata)
           stop
         end if
         if (get_header_r('ARK-GAIN', gain) < 0) then
           print*, 'The gain needs setting for ', ifname
           print*, 'using the program pix_info.'
           deallocate(data,axdata)
           stop
         end if
         jflag=inpflg(naxis, pix_flg)

         if (.not. opened_offsets) then
           print*, ' '
           print*, '* This program also requires a 6 or 2-coeff offsets file.'
           offsets: do
             print*, '> Give this file name (<cr>=offsets.pin).'
             offset=' '
             read (*, '(a)') offset
             if (offset .eq. ' ') offset='offsets.pin'
             inquire(file=offset, exist=there)
             if (.not. there) then
               print*, '* File not found.'
             else
               exit offsets
             end if
           end do offsets
           call arkopn(iunit1, offset, ' ', 'OLD', 'READONLY', 'FORMATTED',' ', 0)
           opened_offsets=.true.
         end if

         rewind(iunit1)
         off_find: do
           read(iunit1,*,iostat=iostat) file, jext
           if (iostat < 0) then
             print*, 'Could not find offsets for ', trim(ifname), &
             ' in ', offset
             stop
           end if
           read(iunit1,*) coff
           if (trim(file)==trim(ifname) .and. iext==jext) exit off_find
         end do off_find
         print*, 'Found offsets of ', coff
         start_ccd=0
         do i=1, nstar_all
           if (star_all(i)%ccd == iext) then
             if (start_ccd == 0) start_ccd=i
             end_ccd=i
             star(i)=star_all(i)
             star(i)%x=star_all(i)%x*coff(1) + star_all(i)%y*coff(2) + coff(3)
             star(i)%y=star_all(i)%y*coff(4) + star_all(i)%x*coff(5) + coff(6)
           end if
         end do

         print*, 'Determining sky and noise levels.'
         call back(data, low, high, ibox, noise_map, median_noise, &
         pix_flg)

         ! S/R back sets the noise very large in bad regions.  This
         ! is fine for sig, since an area with large noise will 
         ! produce no events.  When applying a star list from another
         ! frame in the same filter, it can cause a numerical error.
         good_max=minval(data)
         do iaxis2=low(2), high(2)
           do iaxis1=low(1), high(1)
             if (pix_flg(iaxis1,iaxis2) == 'O') then
               good_max=max(good_max,data(iaxis1,iaxis2))
             end if
           end do
         end do         
         where (noise_map > good_max) noise_map=good_max+1.0

         i=index(ifname, '.') 
         if (i <= 0) i=len_trim(ifname)
         ofname_root=ifname(1:i-1)//'_'
         write(ofname_root(i+1:i+2),'(i2.2)') iext
         call shape_filter(data, star(start_ccd:end_ccd), &
         end_ccd-start_ccd+1, seeing, low, &
         high, noise_map, gain, pix_flg, nx, ny, ofname_root, .false.)

         deallocate(noise_map)
         deallocate(data, axdata, pix_flg)

         ! If the star was saturated or in some other way broken in the long,
         ! take the position from the short.
         find_better: do istar=start_ccd, end_ccd
           if (star_all(istar)%col(1)%flg == 'OO') cycle find_better
           if (star_all(istar)%col(1)%flg == 'ON') cycle find_better
           star_all(istar)%x =   ((star(istar)%x-coff(3)) - &
           (coff(2)/coff(4))*(star(istar)%y-coff(6)))/ &
           (coff(1) - (coff(5)*coff(2)/coff(4)))
           star_all(istar)%y = &
           (star(istar)%y - coff(5)*star_all(istar)%x - coff(6))/coff(4)
           print*, 'Moved a star by ', star(istar)%x-star_all(istar)%x, &
           star(istar)%y-star_all(istar)%y
           star_all(istar)%col(1)%flg=star(istar)%col(1)%flg
         end do find_better

       end do extension

       close(iunit1)

       open(unit=4, file='psf.cat', status='unknown')
       write(4,*) '1 filter'
       write(4,*) '?'
       write(4,*) 'Catalogue by flamsteed, using a seeing of', &
       seeing, ' pixels and a sigma of ', sigma
       open(unit=7, file='stars.cat', status='unknown')
       write(7,*) '1 filter'
       write(7,*) '?'
       write(7,*) 'Catalogue by flamsteed, using a seeing of', &
       seeing, ' pixels and a sigma of ', sigma
       do i=1, nstar_all
         call flux_to_mag(star_all(i)%col(1))
         call write_star(7, star_all(i),1)
         if (star_all(i)%col(1)%flg == 'OO') call write_star(4, star_all(i),1)
       end do 
       close(4)
       close(7)
         
     end program flamsteed
