       program qphot

       ! Does quck aperture photometry for single star work.
        
       use opt_extr
       use ark_file_io  
         
       implicit none

       real, dimension(:,:), allocatable :: image, aximage
       integer, dimension(2) :: naxis, iloc, low
       integer :: iflag
       real, parameter :: fwhm=3.0, adu=6.7, aprad=4.0
       real :: satur
       ! Bit returned by skyfit.
       real :: skycnt, skyerr, skynos
       ! For gfit.
       real, dimension(6) :: a_par
       real, dimension(2) :: e_pos
       ! The answer.
       real :: flux, error
       character(len=50) :: filnam, timfil
       integer :: iostat, iunit2, ntim
       logical :: there
       real, allocatable, dimension(:) :: time, dtime
       character(len=50), allocatable, dimension(:) :: timnam
       real :: timec, dtimec
       integer :: i, j

       open(unit=1, file='qphot.dat', status='new')
       write(1,*)
       write(1,*)
       write(1,*)

       print*, '* We require a file which has the image file'
       print*, '* name on one line, and the observation time and'
       print*, '* and half the exposure length on the next (in any'
       print*, '* units you like, they are just copied to the output).'
       times: do
         print*, &
         '> Give the name of this file ("end"=no file, <cr>=times.dat).'
         timfil=' '
         read (*, '(a)', iostat=iostat) timfil
         if (iostat < 0) exit times
         if (timfil=='END' .or. timfil=='end') exit times
         if (timfil == ' ') timfil='times.dat'
         inquire(file=timfil, exist=there)
         if (there) then
!          Open the file with the times that each image was taken.
           call arkopn(iunit2, timfil, ' ', 'OLD', 'READONLY', &
           'FORMATTED', ' ', 0)
           ntim=0
           cttimes:  do
             read(iunit2, '(a)', iostat=iostat) timfil
             if (iostat < 0) exit cttimes
             read(iunit2, '(a)', iostat=iostat) timfil
             if (iostat < 0) exit cttimes
             ntim=ntim+1
           end do cttimes
           allocate(time(ntim))
           allocate(dtime(ntim))
           allocate(timnam(ntim))
           rewind (iunit2)
           do i=1, ntim
             read(iunit2, '(a)') timnam(i)
             read(iunit2, *) time(i), dtime(i)
!            Remove leading blanks.
              deblank: do j=0, 50
                if (timnam(i)(1:1) .eq. ' ') then
                  timnam(i)=timnam(i)(2:50)
                else
                  exit deblank
                end if
              end do deblank
           end do
           close(iunit2)
           print*, '* ', ntim, ' times read.'
           exit times
         else
           print*, '* File not found.'
         end if
       end do times

       frame: do
         call extark('raw')
         iflag=sizark(naxis)
         if (iflag == -1) exit frame
         allocate (image(naxis(1),naxis(2)))
         allocate (aximage(maxval(naxis),2))
         iflag=inpark(naxis, image, aximage)
         call lstark(filnam)
         low=1
         ! Find the position of the star.
         iloc=maxloc(image)
         ! Find the sky.
         satur=16000.0
         call skyfit(image, real(iloc(1)), real(iloc(2)), fwhm, 30, &
         low, naxis, skycnt, skyerr, skynos, iflag)
         if (iflag /= 0) then
           print*, 'Sky fit failed for file ', filnam
           print*, 'Flag was ', iflag
           deallocate(image, aximage)
           cycle frame
         end if
         ! Now fit the star position.
         a_par(1) = fwhm /1.665
         a_par(2) = fwhm /1.665
         a_par(3) = 0.0
         a_par(4) = image(iloc(1),iloc(2)) - skycnt
         a_par(5) = real(iloc(1))
         a_par(6) = real(iloc(2))
         call gfit(image, .false., 3.0, .false., skycnt, skynos, &
         adu, satur, a_par, e_pos, iflag)
         if (iflag /= 0) then
           print*, 'Star fit failed for file ', filnam
           print*, 'Flag was ', iflag
           deallocate(image, aximage)
           cycle frame
         end if
         satur=16000.0
         flux = sum_aper(image, aprad, skycnt, skyerr, skynos, &
         a_par(5), a_par(6), adu, satur, low, naxis, error, iflag)
         ! Try and match the file name to a known time.
         if (ntim /= 0) then
           do i=1, ntim
             if (filnam .eq. timnam(i)) then
               timec=time(i)
               dtimec=dtime(i)
               goto 560
             end if
           end do
!          When this fails use the frame number.
           print*, 'Failed to find a time to match this frame.'
           print*, 'Will use the frame number.'
         end if       
         do i=1, 50
           if (ichar(filnam(i:i)).lt.48 .or. &
           ichar(filnam(i:i)).gt.57) filnam(i:i)=' '
         end do 
         timec=0.0
         read (filnam, *, iostat=iostat) timec
         dtimec=0.5
560      print*, 'Frame time is ', timec, '+/-', dtimec

         if (iflag >= 0.0) then
           print*, timec, flux, error, dtimec
           write(1,*) timec, flux, error, dtimec
         end if
         deallocate(image, aximage)
       end do frame

       deallocate(time, dtime, timnam)
       close(1)

       end program qphot
