  program pix_info

    ! A general program for putting the header items into a FITS file
    ! so that flamsteed and friends can play with it.

    ! At the moment tested for multi-extension fits files (for which it
    ! works if there are not bad pixel files), and shoudl work for single
    ! fits images.

    use ark_file_io
    use put_header

    implicit none
    
    real, allocatable, dimension(:,:) :: data, axdata
    integer :: iflag
    integer, dimension(2) :: naxis, low, high
    ! The file name.
    character(len=50) :: filnam
    ! The extension number.
    integer :: iext
    real :: satur, gain, ron, lin
    ! Are optional properties there?
    logical :: readout_noise, linear
    character(len=50) :: basic, instr
    character, dimension(:,:), allocatable :: pix_flg
    ! For dealing with bad pixels.
    real, allocatable, dimension(:) :: xbpl, ybpl
    integer :: nbpl, ibpl
    integer :: i

    call setbug()

    print*, '> Give xlow, xhigh, ylow, yhigh (or zeros to use full area).'
    read(*,*) low(1), high(1), low(2), high(2)
    print*, &
    '> Give ADU level up to which detector is linear, or type <cr> of not required.'
    read(*,'(a)') instr
    if (instr /= ' ') then
      read(instr,*) lin
      linear = .true.
    else
      linear = .false.
    end if                              
    print*, '> Give saturation level in ADU (<cr>=65535).'
    read(*,'(a)') instr
    if (instr /= ' ') then
      read(instr,*) satur
    else
      satur=65535.0
    end if                              
    print*, '> Give the gain in electrons per ADU.'
    read(*,*) gain
    instr=' '
    print*, '> Give the readout noise in ADU or type <cr> if not required.'
    read(*,'(a)') instr
    if (instr /= ' ') then
      read(instr,*) ron
      readout_noise = .true.
    else
      readout_noise = .false.
    end if
      
    basic=' '
    print*, '  The program can take an initial pixel flag map, and ammend it.'
    print*, ' '
    print*, '> If you have such a map, give its name, otherwise hit <cr>.'
    read(*,'(a)')  basic

    print*, '  You can also supply a file of bad pixels as a list of X-Y'
    print*, '  points, one per line (with the first 3 lines ignored).'
    print*, '  Give the name of this file (or type end if there isn''t one).'
    ! Read the bad pixel file.
    ibpl=inpark(nbpl, ybpl, xbpl)
    if (ibpl < 0) nbpl=0
      
    do

      call extark("raw")
      iflag=inpark(naxis, data, axdata)
      if (iflag < 0) exit
      call lstark_in(filnam, iext)

      if (naxis(1)*naxis(2) > 0) then
        ! Add header items.
        ! Set the edges of the area.
        if (low(1) == 0) then
          call put_header_i('ARK-LO1', 1,  'AXIS1 LOWEST LIT PIXEL',  1)
        else
          call put_header_i('ARK-LO1', low(1),  'AXIS1 LOWEST LIT PIXEL',  1)
        end if
        if (high(1) == 0) then
          call put_header_i('ARK-HI1', naxis(1), 'AXIS1 HIGHEST LIT PIXEL', 1)
        else
          call put_header_i('ARK-HI1', high(1), 'AXIS1 HIGHEST LIT PIXEL', 1)
        end if
        if (low(2) == 0) then
          call put_header_i('ARK-LO2', 1,  'AXIS2 LOWEST LIT PIXEL',  1)
        else
          call put_header_i('ARK-LO2', low(2),  'AXIS2 LOWEST LIT PIXEL',  1)
        end if
        if (high(1) == 0) then
          call put_header_i('ARK-HI2', naxis(2), 'AXIS2 HIGHEST LIT PIXEL', 1)
        else
          call put_header_i('ARK-HI2', high(2), 'AXIS2 HIGHEST LIT PIXEL', 1)
        end if
        call put_header_r('ARK-GAIN', gain, 'GAIN IN ELECTRONS/ADU',1)
        ! Now the readout noise.
        if (readout_noise) &
        call put_header_r('ARK-RON', ron, 'READOUT NOISE IN ADU',1)
        if (basic /= ' ') then
          iflag=inpflg(naxis, pix_flg, basic)
          if (iflag == 0) then
            print*, 'Cannot find file ', basic
            stop
          end if
        else
          allocate(pix_flg(naxis(1),naxis(2)))
          pix_flg='O'
        end if
      
        ! The order here has been considered carefully.

        ! Set the edge area, overwriting all other flags.
        if (low(1)  > 1       .and. low(1) /=0) pix_flg(1:low(1)-1,:        )='E'
        if (low(2)  > 1       .and. low(2) /=0) pix_flg(:,1:low(1)-1        )='E'
        if (high(1) < naxis(1).and. high(1)/=0) pix_flg(high(1)+1:naxis(1),:)='E'
        if (high(2) < naxis(2).and. high(2)/=0) pix_flg(:,high(2)+1:naxis(2))='E'
      
        ! The bad pixels.  Only bad pixel things not already flagged.
        if (nbpl > 0) then
          do ibpl=1, nbpl
            if (pix_flg(nint(xbpl(ibpl)), nint(ybpl(ibpl))) == 'O') &
            pix_flg(nint(xbpl(ibpl)), nint(ybpl(ibpl)))='F'
          end do
        end if

        ! First do saturation and linearity.  These will not be set if the
        ! pixel is bad or non linear.
        where(data>=satur .and. pix_flg=='O') pix_flg='S'
        if (linear) where(data>=lin .and. pix_flg=='O') pix_flg='L'

        ! Reset the axis scales.  Most programs don't sue these, but
        ! the ones which do therefore cause confusion.  So, set the
        ! scales to be pixel numbers.
        do i=1, naxis(1)
          axdata(i,1)=real(i)
        end do
        do i=1, naxis(2)
          axdata(i,2)=real(i)
        end do

      end if

      ! Write out the data and flag file.
      filnam(1:1)='h'
      call nxtark_out(filnam, iext)
      iflag=makflg(naxis, pix_flg)
      iflag=makark(naxis, data, axdata)

      if (naxis(1)*naxis(2) > 0 ) deallocate(data, axdata, pix_flg)

    end do

    if (allocated(xbpl)) deallocate(xbpl)
    if (allocated(ybpl)) deallocate(ybpl)

  end program pix_info
