## This routine requires a little thought before recompiling.  
## First it should use make use of makark rather that wrfits.
## Also, Alasdair thinks there is a bug, and the definition of
## l_Par should read l_Par( m_Par, 2) as it is later referenced
## as having two elements in its second dimension.


	Subroutine GRID_2D( X, Y, W, Points, mode_W, A, dA, n_Par, n_Term, &
      				d_Chi, sig_Clip, m_Rep, FUNCTN, FDERIV,    &
     				g_Chi, Dim_1, Dim_2, g_Par, g_Min, g_Max,  &
     							out_Unit, of_Name )

!	2D-Grid search in order to derive the actual errors on the fitted
!	Parameters
!	WARNING:: Can cause errors in case of a mixed-mode-linking!

!	Version 2.0
!	Output data in ARK-FITS format; adjusted for use in new RVFIT
!							KM, September 19, 1989

	Integer m_Par, m_Header
	Parameter( m_Par = 32 )
	Parameter( m_Header = 360 )
!					INPUT/OUTPUT PARAMETERS
	Integer Points			! Number of data points
	Real X( Points )		! X data
	Real Y( Points )		! Y data - measurments
	Real W( Points )		! Weight array for above
	Integer mode_W			! 1 if Poissonian, 0 for equal
	Integer n_Par			! Number of parameters
	Real A( n_Par )			! The starting parameters
	Real dA( n_Par )		! For numerical differentiation
	Integer n_Term			! Number of free parameters
	Real d_Chi			! Fractional improvement in Chi
	Real sig_Clip			! Outlier rejection threshold
	Integer m_Rep			! Maximum repetition of fits
	Real FUNCTN			! Model function
	External FUNCTN			! to be supplied by the use
	External FDERIV			! together with derivatives
	Integer Dim_1, Dim_2		! Size of grid search
	Real g_Chi( Dim_1, Dim_2 )	! Chi-squared grid
	Integer g_Par( 2 )		! Parameters to be grid-searched
	Real g_Min( 2 ), g_Max( 2 )	! Limit on parameters
	Integer out_Unit		! Output unit, if non-zero
	Character*50 of_Name		! Output file name
!					OTHER VARIABLES
	Integer temp_TF( m_Par )	! Save Table_Free
	Integer Linked( 2 )		! Number of para's linked with Par's
	Integer l_Par( m_Par, 1 )	! List of above
	Integer Host( 2 )		! The master parameter of above
	Real Off( m_Par, 2 )		! List of offset for above
	Integer Rest			! Number of para's not linked to either
	Integer list_rest( m_Par )	! List of above
	Real Step_1, Add_1		! Grid search steps for Par( 1 )
	Real Step_2, Add_2		! Grid search steps for Par( 2 )
	Real g_A( m_Par )		! Temporary para's
	Real Covar( m_Par, m_Par )	! Covariant matrix - ignore
	Real dY				! Data - FUNCTN
	Real min_Chi			! Minimum chi-squared
	Integer Flag			! CURFIT flag -- ignore for now
	Integer I, J, K, K2, L, M		! Dummy variables
	Integer save_Term
	Integer Unit
        Integer iflag

	Integer Index( m_Par )
	Integer Table_Free( m_Par )
	Integer l_Index( m_Par )
	Real l_Offset( m_Par )
	Character*80 Header( m_Header )
	Integer n_Header
	Common / free_list / Index, Table_Free
	Common / link_status / l_Index, l_Offset
	Common / fits_Head / Header, n_Header
 
!		First, check the chosen parameters are not fixed
!		nor linked to each other...if so, reject
	If( l_Index( g_Par( 1 ) ) .eq. l_Index( g_Par( 2 ) ) ) Then
	  Write( *, 110 ) g_Par( 1 ), g_Par( 2 )
110	  Format( ' ERROR:: Parameters #', I2, ' and #', I2, ' are linked' / &
      			'         This call to GRID_2D is aborted' )
	  Return
	Else If( Table_Free( g_Par( 1 ) ) .eq. 1 ) Then
	  Write( *, 120 ) g_Par( 1 )
120	  Format( ' ERROR:: Parameter #', I2, ' is fixed' / &
      			'         This call to GRID_2D is aborted' )
	  Return
	Else If( Table_Free( g_Par( 2 ) ) .eq. 1 ) Then
	  Write( *, 120 ) g_Par( 2 )
	  Return
	End If
!		Next, deal with linked parameters
	Rest = n_Par
	Do K = 1, Rest
	  list_Rest( K ) = K
	End Do
	Do K = 1, n_Par
	  temp_TF( K ) = Table_Free( K )
	End Do
	Do L = 1, 2
	  Do K2 = 1, Rest
	    If( list_Rest( K2 ) .eq. g_Par( L ) ) Goto 130
	  End Do
130	  Continue
	  Rest = Rest - 1
	  Do K = K2, Rest
	    list_Rest( K ) = list_Rest( K + 1 )
	  End Do
	  Host( L ) = l_Index( g_Par( L ) )
	  Linked( L ) = 0
	  Do K = 1, n_Par
	    If( l_Index( K ) .eq. Host( L ) .and. K .ne. g_Par( L ) ) Then
	      Linked( L ) = Linked( L ) + 1
	      l_Par( Linked( L ), L ) = K
	      Do K2 = 1, Rest
		If( list_Rest( K2 ) .eq. K ) Goto 140
	      End Do
140	      Continue
	      Rest = Rest - 1
	      Do M = K2, Rest
		list_Rest( M ) = list_Rest( M + 1 )
	      End Do
	      If( Host( L ) .eq. g_Par( L ) ) Then
		Off( K, L ) = l_Offset( K )
	      Else
		If( temp_TF( K ) .eq. -1 ) Then
		  Off( K, L ) = l_Offset( K ) - l_Offset( g_Par( L ) )
		Else
		  Off( K, L ) = l_Offset( K ) / l_Offset( g_Par( L ) )
		End If
	      End If
	      If( K .ne. Host( L ) ) Then
		iflag = SEPARATE_PAR( K, n_Par, n_Term )
		iflag = FIX_PAR( K, n_Par, n_Term )
	      End If
	    End If
	  End Do
	  If( g_Par( L ) .ne. Host( L ) ) Then
	    iflag = SEPARATE_PAR( g_Par( L ), n_Par, n_Term )
	    iflag = FIX_PAR( Host( L ), n_Par, n_Term )
	  End If
	  iflag = FIX_PAR( g_Par( L ), n_Par, n_Term )
	End Do
	save_Term = n_Term		! Save the n_Term during gridding
!		Calculate the weight array, steps etc.
	If( mode_W .eq. 1 ) Then	! Might as well calculate now
	  Do J = 1, Points
	    If( Y( J ) .gt. 0.0 ) Then
	      W( J ) = 1.0 / Y( J )
	    Else
	      W( J ) = 0.0
	    End If
	  End Do
	Else If( mode_W .eq. 0 ) Then
	  Do J = 1, Points
	    W( J ) = 1.0
	  End Do
	End If
	Step_1 = ( g_Max( 1 ) - g_Min( 1 ) ) / ( Dim_1 - 1 )
	Add_1 = g_Min( 1 ) - Step_1
	Step_2 = ( g_Max( 2 ) - g_Min( 2 ) ) / ( Dim_2 - 1 )
	Add_2 = g_Min( 2 ) - Step_2
!		Now loop through the grid points
	Print *
	min_Chi = 1.0E+33
	Do I = 1, Dim_1
	  g_A( g_Par( 1 ) ) = I * Step_1 + Add_1
	  Do K = 1, Linked( 1 )	! don't forget the linked ones
	    L = l_Par( K, 1 )
	    If( temp_TF( L ) .eq. -1 ) Then
	      g_A( L ) = g_A( g_Par( 1 ) ) + Off( K, 1 )
	    Else
	      g_A( L ) = g_A( g_Par( 1 ) ) * Off( K, 1 )
	    End If
	  End Do
	  Do J = 1, Dim_2
	    g_A( g_Par( 2 ) ) = J * Step_2 + Add_2
	    Do K = 1, Linked( 2 )
	      L = l_Par( K, 2 )
	      If( temp_TF( L ) .eq. -1 ) Then
		g_A( L ) = g_A( g_Par( 2 ) ) + Off( K, 2 )
	      Else
		g_A( L ) = g_A( g_Par( 2 ) ) * Off( K, 2 )
	      End If
	    End Do
	    Do K = 1, Rest
	      g_A( list_Rest( K ) ) = A( list_Rest( K ) )
	    End Do
	    If( n_Term .ge. 1 ) Then
	      Flag = 1024
	      Call CURFIT( X, Y, W, Points, -1, g_A, dA, n_Par, &
      				n_Term, Covar, m_Par, d_Chi, g_Chi( I, J ), &
      				sig_Clip, m_Rep, Flag, FUNCTN, FDERIV )
	    Else
	      g_Chi( I, J ) = 0.0
	      Do K = 1, Points
		dY = Y( K ) - FUNCTN( X( K ), g_A, n_Par )
		g_Chi( I, J ) = g_Chi( I, J ) + dY * dY * W( K )
	      End Do
	    End If
	    min_Chi = Min( min_Chi, g_Chi( I, J ) )
	  End Do
	  Write( *, 210 ) I
210	  Format( '+* Finished ', I2, 'th column/row' )
	End Do
!		Recover the initial status
	Do L = 1, 2
	  iflag = FREE_PAR( g_Par( L ), n_Par, n_Term )
	  If( g_Par( L ) .ne. Host( L ) ) Then
	    iflag = FREE_PAR( Host( L ), n_Par, n_Term )
	    If( temp_TF( g_Par( L ) ) .eq. -1 ) Then
	      iflag = LINK_PAR( Host( L ), g_Par( L ), l_Offset( g_Par( L ) ), &
      							0, A, n_Par, n_Term )
	    Else
	      iflag = LINK_PAR( Host( L ), g_Par( L ), l_Offset( g_Par( L ) ), &
      							1, A, n_Par, n_Term )
	    End If
	  End If
	  Do K = 1, Linked( L )
	    M = l_Par( K, L )
	    If( M .ne. Host( L ) ) Then
	      iflag = FREE_PAR( M, n_Par, n_Term )
	      If( temp_TF( M ) .eq. -1 ) Then
		iflag = LINK_PAR &
      			( Host( L ), M, l_Offset( M ), 0, A, n_Par, n_Term )
	      Else
		iflag = LINK_PAR &
      			( Host( L ), M, l_Offset( M ), 1, A, n_Par, n_Term )
	      End If
	    End If
	  End Do
	End Do

	If( out_Unit .lt. 0 ) Then	! Rescale to 1.0
	  Do I = 1, Dim_1
	    Do J = 1, Dim_2
	      g_Chi( I, J ) = g_Chi( I, J ) / min_Chi
	    End Do
	  End Do
	  Unit = -out_Unit
	Else
	  Unit = out_Unit
	End If
!		Output CHIGRID/RVLMT compatible file, if desired
	If( Unit .gt. 0 ) Then
	  Header( 1 ) = 'PROGRAM = ''GRID-2D''            /FILE WRITING PROGRAM'
	  Header( 2 ) = &
      		'CRPIX1  =                    1 /CRVAL1 REFERS TO FIRST PIXEL'
	  Header( 3 ) = &
      		'CRVAL1  =                      /VALUE OF FIRST AXIS1 PIXEL'
	  Write( Header( 3 )( 19: 30 ), '(1PE12.5)' ) g_Min( 1 )
	  Header( 4 ) = 'CDELT1  =                      /STEP FOR AXIS1 PIXELS'
	  Write( Header( 4 )( 19: 30 ), '(1PE12.5)' ) Step_1
	  Header( 5 ) = 'GRID-ID1=                      /LABEL OF AXIS1'
	  Write( Header( 5 )( 29: 30 ), '(I2)' ) g_Par( 1 )
	  Header( 6 ) = &
      		'CRPIX2  =                    1 /CRVAL2 REFERS TO FIRST PIXEL'
	  Header( 7 ) = &
      		'CRVAL2  =                      /VALUE OF FIRST AXIS2 PIXEL'
	  Write( Header( 7 )( 19: 30 ), '(1PE12.5)' ) g_Min( 2 )
	  Header( 8 ) = 'CDELT2  =                      /STEP FOR AXIS2 PIXELS'
	  Write( Header( 8 )( 19: 30 ), '(1PE12.5)' ) Step_2
	  Header( 9 ) = 'GRID-ID2=                      /LABEL OF AXIS2'
	  Write( Header( 9 )( 29: 30 ), '(I2)' ) g_Par( 2 )
	  Header( 10 ) = &
      		'NP-FIT  =                      /NUMBER OF DATAPOINTS FOR FIT'
	  Write( Header( 10 )( 26: 30 ), '(I5)' ) Points
	  Header( 11 ) = &
      		'DOF-FIT =                      /DEGREE OF FREEDOM OF FIT'
	  Write( Header( 11 )( 26: 30 ), '(I5)' ) save_Term
	  n_Header = 11
	  Flag = 0
	  Call WRFITS( g_Chi, Dim_1, Dim_2, 1, Dim_1, 1, Dim_2, Unit, of_Name, Flag )
	End If

	End Subroutine GRID_2D
